﻿Imports System.Collections.ObjectModel
Imports System.Drawing
Imports System.Drawing.Printing
Imports System.IO
Imports System.Windows.Forms
Imports MapUtilities.My.Resources.PageLayoutControlResources
Imports System.ComponentModel
Imports MapUtilities.CustomLegendItem.EnumLegendItemType
Imports System.Runtime.Serialization.Formatters.Binary
Imports System.Security.AccessControl
Imports ErrHandler = CentralExceptionManager.HandledExceptionManager
Imports MapUtilities.frmPrintToPdf.EnumPdfFormatOption
Imports MapUtilities.frmPrintToPdf.EnumWriteGeoReferenceOption
Imports MapUtilities.frmPrintToPdf
Imports MapUtilities.PdfGeoRef
Imports MapUtilities.UnmanagedAPI
Imports ThinkGeo.MapSuite.WinForms
Imports ThinkGeo.MapSuite.Layers
Imports ThinkGeo.MapSuite
Imports ThinkGeo.MapSuite.Drawing
Imports ThinkGeo.MapSuite.Shapes
Imports ThinkGeo.MapSuite.Styles
Imports ThinkGeo.MapSuite.Serialize

Public Class SilvicsPageLayout
	Inherits WinformsMap
	Private moSelectedElements As GeoCollection(Of PrinterLayer)
	Private components As System.ComponentModel.IContainer
	Private WithEvents moPrintDocument As PrintDocument = New PrintDocument()
	Public Const csLegendLayerBaseKey As String = "LegendLayer"
	Public Const csMagneticNorthArrowBaseKey = "MagneticNorthArrow"
	Public Const csScaleLineLayerBaseKey As String = "ScaleLineLayer"
	Public Const csScaleTextLayerBaseKey As String = "ScaleTextLayer"
	Public Const csDataGridBaseKey As String = "DataGridLayer"
	Public Const csScaleBarLayerBaseKey As String = "ScaleBarLayer"
	Public Const csPrinterOverlayBaseKey As String = "PrinterOverlay"
	Public Const csMapPrinterLayerBaseKey As String = "MapPrinterLayer"
	Public Const csPageLayerKey As String = "PageLayer"
	Private miMapNumber As Integer = 0
	Private moSelectedPrinterLayer As PrinterLayer
	Public Property TemplateFileName As String = String.Empty
	Private moActiveLayoutMap As CustomMapPrinterLayer
	Private msngDpi As Single
	Private moGraphics As Graphics = Me.CreateGraphics
	Public Property HighlightSelectedFeatures As Boolean = False
	Private moPrinterOverlays As New GeoCollection(Of PrinterInteractiveOverLay)
	Private moProgressForm As frmProgress
	Private WithEvents moPrinterWorker As BackgroundWorker
	Private meCurrentActiveView As Enums.EnumActiveView = Enums.EnumActiveView.Unknown
	Private meCurrentMapUnit As GeographyUnit
	Private WithEvents moCurPrinterOverlay As CustomPrinterOverlay = Nothing
	Private WithEvents moExtentOverlay As ExtentInteractiveOverlay = Me.ExtentOverlay
	Public Property HasSyncedToParentMap As Boolean = False
	Public Property StandardLayer As StandardLayer
	Public Enum EnumScaleMode
		SpecificScale = 1
		BestFit = 2
	End Enum
	Public ReadOnly Property PrinterOverlays As GeoCollection(Of PrinterInteractiveOverLay)
		Get
			Return moPrinterOverlays
		End Get
	End Property
	Public Enum EnumBuildLegendOption
		CreateDefaultLegend = 1
		CopyExistingLegend = 2
	End Enum
	Public Property ActiveLayoutMap As CustomMapPrinterLayer
		Get
			If moActiveLayoutMap Is Nothing Then
				SetActiveLayoutMap()
			End If
			Return moActiveLayoutMap
		End Get
		Set(value As CustomMapPrinterLayer)
			moActiveLayoutMap = value
		End Set
	End Property
	Public Sub SetActiveLayoutMap()
		For Each oPrinterLayer As PrinterLayer In PrinterOverlay.PrinterLayers
			If TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
				moActiveLayoutMap = CType(oPrinterLayer, CustomMapPrinterLayer)
				Exit For
			End If
		Next oPrinterLayer
	End Sub
	Public ReadOnly Property CurrentParentMap As CustomWinformsMap
		Get
			Return moActiveLayoutMap.ParentMap
		End Get
	End Property
	Public Enum EnumLegendSyncBehavior
		SyncLegend = 1
		DoNotSyncLegend = 2
	End Enum
	Public Enum EnumPageLayoutMapRedrawMode
		Continuous = 1
		OnDemand = 2
	End Enum
	Public Enum EnumElementOrderType
		SendToBack = 1
		BringToFront = 2
	End Enum
	Public Enum EnumCenterType
		Horizontal = 1
		Vertical = 2
	End Enum
	Public Enum EnumPagePosition
		TopLeft = 1
		TopRight = 2
		LowerLeft = 3
		LowerRight = 4
		Current = 5
	End Enum

	Public Sub New()
		MyBase.New()
		InitializeComponent()
		Me.DrawingQuality = ThinkGeo.MapSuite.Drawing.DrawingQuality.CanvasSettings
		msngDpi = moGraphics.DpiX

		'***** Setup the map unit, you want to always use feet or meters for printer layout maps
		MapUnit = GeographyUnit.Feet

		ZoomLevelSnapping = ZoomLevelSnappingMode.SnapUp

		'***** Create the default zoom levels for the layout map
		ZoomLevelSet = New PrinterZoomLevelSet(MapUnit, PrinterHelper.GetPointsPerGeographyUnit(MapUnit))
		ZoomLevelHelper.SetPageLayoutZoomLevels(CType(ZoomLevelSet, PrinterZoomLevelSet), msngDpi)
		'MaximumScale = CInt(ZoomLevelSet.ZoomLevel01.Scale)
		MaximumScale = CInt(ZoomLevelSet.CustomZoomLevels(0).Scale)
		'**** Only allow the page layout to zoom into ZoomLevel19. If zoomed to ZoomLevel20 an out
		'***** of memory exception frequently occurs.
		'***** SET 11/17/2015 - 7:04 AM
		'***** QB #
		'MinimumScale = CInt(ZoomLevelSet.ZoomLevel19.Scale)
		MinimumScale = CInt(ZoomLevelSet.CustomZoomLevels(ZoomLevelSet.CustomZoomLevels.Count - 1).Scale)
		'***** Set the background color to gray so there is contrast with the white page
		BackgroundOverlay.BackgroundBrush = New GeoSolidBrush(GeoColor.StandardColors.LightGray)

	End Sub

	Public Property ActiveMapUnit() As GeographyUnit
		Get
			Dim eUnit As GeographyUnit = GeographyUnit.Unknown
			If Not moActiveLayoutMap Is Nothing Then
				eUnit = moActiveLayoutMap.MapUnit
			End If
			Return eUnit
		End Get
		Set(ByVal value As GeographyUnit)
			If Not moActiveLayoutMap Is Nothing Then
				moActiveLayoutMap.MapUnit = value
			End If

			'***** Find any ScaleLines or ScaleBars
			'***** and set their MapUnit
			mSetScaleBarAndLineMapUnits(value)
			mSetMagneticNorthArrowProj()
		End Set
	End Property
	Public Sub AddMapLayer(ByVal oLayerToAdd As Layer, ByVal oMap As CustomWinformsMap, ByVal eSyncLegend As EnumLegendSyncBehavior)
		Dim iIndex As Integer
		Dim oPageLayoutMap As CustomMapPrinterLayer = ActiveLayoutMap
		'***** The layer has aleady been added to the 'main' map.
		'***** Now it needs to be added to the layout map

		'***** Now let's find the index where the new layer should be added
		iIndex = mFindLayerIndex(oLayerToAdd, oMap)

		oPageLayoutMap.Layers.Insert(iIndex, oLayerToAdd)
		SyncLegendToLayers(oMap, oLayerToAdd)
	End Sub
	Public Sub RemoveMapLayer(oMap As CustomWinformsMap, ByVal oLayerToRemove As Layer, ByVal eSyncLegend As EnumLegendSyncBehavior)
		Dim oPageLayoutMap As CustomMapPrinterLayer = mLayoutMapFromParentMap(oMap)
		'***** The layer has already been removed from the 'main' map
		'***** and now needs to be removed from the layout map
		With oPageLayoutMap
			If TypeOf oLayerToRemove Is GroupLayer Then
				'***** Call the sub recursively to remove all the layers contained
				'***** by this group layer
				For Each oLayer As Layer In CType(oLayerToRemove, GroupLayer).Layers
					RemoveMapLayer(oMap, oLayer, eSyncLegend)
				Next oLayer
			Else
				'***** Remove the layer from the PageLayout map
				If .Layers.Contains(oLayerToRemove) Then
					.Layers.Remove(oLayerToRemove)
				End If
			End If
		End With
		If eSyncLegend = EnumLegendSyncBehavior.SyncLegend Then
			SyncLegendToLayers(oMap)
		End If
	End Sub
	Private Function mLayoutMapFromParentMap(oMap As CustomWinformsMap) As CustomMapPrinterLayer
		Dim sKey As String = mGetPageLayoutKeyFromElementKey(csMapPrinterLayerBaseKey & "_" & oMap.PageLayoutMapNumber.ToString)
		Return CType(mGetPrinterOverlay.PrinterLayers(sKey), CustomMapPrinterLayer)
	End Function
	Public Sub RepositionMapLayer(oLayerToMove As Layer, oMap As CustomWinformsMap)
		'***** With this type of layers collection we cannot move the layer,
		'***** we have to remove it from the collection and add it back at
		'***** its new index
		Dim oPageLayoutMap As CustomMapPrinterLayer = ActiveLayoutMap
		If TypeOf oLayerToMove Is GroupLayer Then
			'***** Just empty and refill the page layout map's layers collection
			Dim oLayers As Collection(Of Layer) = oMap.GetAllNonGroupLayers()
			oPageLayoutMap.Layers.Clear()
			For Each oLayer In oLayers
				oPageLayoutMap.Layers.Add(oLayer)
			Next oLayer
		Else
			Dim iNewIndex As Integer = mFindLayerIndex(oLayerToMove, oMap)
			RemoveMapLayer(oMap, oLayerToMove, EnumLegendSyncBehavior.DoNotSyncLegend)
			oPageLayoutMap.Layers.Insert(iNewIndex, oLayerToMove)
		End If
	End Sub
	Private Function mFindLayerIndex(oLayerToFind As Layer, oMap As CustomWinformsMap) As Integer
		Dim iIndex As Integer = 0
		Dim oLayers As GeoCollection(Of Layer) = oMap.GetAllNonGroupLayers()

		For Each oLayer As Layer In oLayers
			If oLayerToFind.Equals(oLayer) Then
				'***** We have found the layer in the 'main' map and we have an index it
				Exit For
			End If
			iIndex += 1
		Next oLayer

		Return iIndex
	End Function
	Public Sub RenameLayerInLegend(oLayerToRename As Layer)
		Dim oItem As CustomLegendItem = Nothing
		Dim bSuccess As Boolean = False
		Dim oLegend As CustomLegendLayer

		If Not ActiveLayoutMap.Legend Is Nothing Then
			oLegend = ActiveLayoutMap.Legend
			If Not TypeOf oLayerToRename Is GroupLayer Then
				For Each oItem In oLegend.LegendItems
					If Not oItem.Layer Is Nothing Then
						If oItem.Layer.Equals(oLayerToRename) Then
							oItem.TextStyle.TextColumnName = oLayerToRename.Name
							bSuccess = True
							Exit For
						End If
					End If
				Next oItem
			End If

			If bSuccess Then
				'***** The layer was re-named in the legend so now we need to 
				'***** see if the width of the new name is too wide to be 
				'***** contained in the legend and widen it if it is.
				BuildLegend(EnumBuildLegendOption.CopyExistingLegend)
			End If
			Me.Refresh()
		End If
	End Sub
	Public Sub mSetElementOrder(ByVal oPrinterLayer As PrinterLayer, ByVal eOrderType As EnumElementOrderType)
		Dim oPrinterOverlay As PrinterInteractiveOverlay = mGetPrinterOverlay()

		If eOrderType = EnumElementOrderType.BringToFront Then
			oPrinterOverlay.PrinterLayers.MoveToTop(oPrinterLayer)
		Else
			oPrinterOverlay.PrinterLayers.MoveTo(oPrinterLayer, 1)
		End If

		Refresh()
	End Sub
	Private Sub mSetScaleBarAndLineMapUnits(eMapUnit As GeographyUnit)
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oPrinterLayer As PrinterLayer

		If Not oPrinterOverlay Is Nothing Then
			For Each oPrinterLayer In oPrinterOverlay.PrinterLayers
				If TypeOf oPrinterLayer Is ScaleLinePrinterLayer Then
					If CType(oPrinterLayer, ScaleLinePrinterLayer).MapPrinterLayer.Name = ActiveLayoutMap.Name Then
						CType(oPrinterLayer, ScaleLinePrinterLayer).MapUnit = eMapUnit
					End If
				ElseIf TypeOf oPrinterLayer Is ScaleBarPrinterLayer Then
					If CType(oPrinterLayer, ScaleBarPrinterLayer).MapPrinterLayer.Name = ActiveLayoutMap.Name Then
						CType(oPrinterLayer, ScaleBarPrinterLayer).MapUnit = eMapUnit
					End If

				End If
			Next oPrinterLayer
		End If
	End Sub
	Private Sub mSetMagneticNorthArrowProj()
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oPrinterLayer As PrinterLayer

		If Not oPrinterOverlay Is Nothing Then
			For Each oPrinterLayer In oPrinterOverlay.PrinterLayers
				If TypeOf oPrinterLayer Is CustomMagDecNorthArrow Then
					CType(oPrinterLayer, CustomMagDecNorthArrow).SetCoordinateSystem()
				End If
			Next oPrinterLayer
		End If
	End Sub
	Public Sub ClearPageLayout()
		For Each oPrinterOverlay In moPrinterOverlays
			oPrinterOverlay.PrinterLayers.Clear()
			If Me.InteractiveOverlays.Contains(oPrinterOverlay.Name) Then
				Me.InteractiveOverlays.Remove(oPrinterOverlay.Name)
			End If
		Next oPrinterOverlay
		moPrinterOverlays.Clear()
		PageLayoutControl.cboPageNumber.Items.Clear()
		PageLayoutControl.cboPageNumber.DataSource = Nothing

	End Sub
	Public Sub CreateDefaultLayout()
		Dim oExtentRectangle As RectangleShape

		Dim oPrinterOverlay As New CustomPrinterOverlay()
		AddHandler oPrinterOverlay.MapKeyUp, AddressOf moPrinterOverlay_MapKeyUp

		'***** Create the PrinterInteractiveOverlay to contain all of the PrinterLayers.
		'***** The interactive overlay allows the layers to be interacted with
		Dim sKey As String = csPrinterOverlayBaseKey & "_" & (moPrinterOverlays.Count + 1).ToString
		oPrinterOverlay.Name = sKey
		InteractiveOverlays.Add(sKey, oPrinterOverlay)
		InteractiveOverlays.MoveToBottom(sKey)
		PrinterOverlays.Add(sKey, oPrinterOverlay)
		PageLayoutControl.AddPrinterOverlayToPageNumberComboBox()

		'***** Create the CustomPagePrinterLayer which shows the page boundary and 
		'***** is the area the user will arrange all of the other layers on top of.
		If LayoutMode <> MapUtilities.PageLayoutControl.EnumPageLayoutMode.AutoLayout Then
			Dim oPrinterPage = New CustomPagePrinterLayer(PrinterPageSize.AnsiA, PrinterOrientation.Portrait)
			With oPrinterPage
				.Open()
				.BackgroundMask = DisplayUtilities.CreateSimpleAreaStyle(GeoColor.StandardColors.White, GeoColor.SimpleColors.Black, 1)
				oPrinterOverlay.PrinterLayers.Add(csPageLayerKey, oPrinterPage)
				'***** Get the page's extent, scale it up by 10%
				oExtentRectangle = CType(RectangleShape.ScaleUp(.GetPosition(), 10), RectangleShape)
				'***** Set oExtentRectangle as the default current extent
				CurrentExtent = oExtentRectangle
			End With
			PageLayoutControl.SetCombosFromPageLayout()
		End If

	End Sub
	Private Sub mAddCenteredTitleLabel()
		Dim oLabelPrinterLayer As LabelPrinterLayer = New LabelPrinterLayer()
		Dim oPageBoundingbox As RectangleShape

		'***** Setup the text and the font..
		With oLabelPrinterLayer
			.Text = "Page Layout"
			.Font = New GeoFont("Arial", 10, DrawingFontStyles.Bold)
			.TextBrush = New GeoSolidBrush(GeoColor.StandardColors.Black)

			'***** Set the labels position so that is it centered on the page one inch from the top                        
			oPageBoundingbox = GetPageBoundingBox(PrintingUnit.Inch)
			Dim oLabelCenter As PointShape = New PointShape()
			oLabelCenter.X = oPageBoundingbox.UpperRightPoint.X - oPageBoundingbox.Width / 2
			oLabelCenter.Y = oPageBoundingbox.UpperLeftPoint.Y - 0.75

			.PrinterWrapMode = PrinterWrapMode.AutoSizeText
			.ResizeMode = PrinterResizeMode.MaintainAspectRatio
			.SetPosition(5, 0.5, oLabelCenter, PrintingUnit.Inch)
		End With
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		oPrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oLabelPrinterLayer)

	End Sub
	Public Sub mCenterElement(ByVal eCenterType As EnumCenterType, ByVal oPrinterLayer As PrinterLayer)
		Dim oNewPrinterLayer As PrinterLayer = Nothing
		Dim oCurPosition As RectangleShape = Nothing
		Dim oNewCenterPoint As PointShape
		Dim iIndex As Integer
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

		iIndex = oPrinterOverlay.PrinterLayers.IndexOf(oPrinterLayer)
		oNewPrinterLayer = oPrinterOverlay.PrinterLayers(iIndex)
		''moPrinterOverlay.PrinterLayers.Remove(oPrinterLayer)
		oPrinterLayer.Open()
		oCurPosition = oPrinterLayer.GetPosition(PrintingUnit.Inch)

		If eCenterType = EnumCenterType.Horizontal Then
			oNewCenterPoint = New PointShape(0, oCurPosition.GetCenterPoint.Y)
		Else
			oNewCenterPoint = New PointShape(oCurPosition.GetCenterPoint.X, 0)
		End If

		oNewPrinterLayer.SetPosition(oCurPosition.Width, oCurPosition.Height, oNewCenterPoint, PrintingUnit.Inch)

		Refresh()

	End Sub
	Private Function mCopyImagePrinterLayer(ByVal oImagePrinterLayer As ImagePrinterLayer) As ImagePrinterLayer
		Dim oNewImagePrinterLayer As ImagePrinterLayer

		With oImagePrinterLayer
			oNewImagePrinterLayer = New ImagePrinterLayer(.Image)
			oNewImagePrinterLayer.BackgroundMask = .BackgroundMask
			oNewImagePrinterLayer.ResizeMode = .ResizeMode
			oNewImagePrinterLayer.DragMode = .DragMode
		End With
		Return oNewImagePrinterLayer
	End Function
	Private Function moCopyLabelPrinterLayer(ByVal oLabelPrinterLayer As LabelPrinterLayer) As LabelPrinterLayer
		Dim oNewLabelPrinterLayer As LabelPrinterLayer

		With oLabelPrinterLayer
			oNewLabelPrinterLayer = New LabelPrinterLayer(.Text, .Font, .TextBrush)
			oNewLabelPrinterLayer.DragMode = .DragMode
			oNewLabelPrinterLayer.ResizeMode = .ResizeMode
			oNewLabelPrinterLayer.BackgroundMask = .BackgroundMask
		End With

		Return oNewLabelPrinterLayer
	End Function
	Private Sub mCopyMap(oMap As CustomWinformsMap, oMapPrinterLayer As CustomMapPrinterLayer)
		Dim oPageLayoutMap As CustomMapPrinterLayer = ActiveLayoutMap
		With oMapPrinterLayer
			oPageLayoutMap = New CustomMapPrinterLayer(.Layers, .MapExtent, .MapUnit)
			oPageLayoutMap.BackgroundMask = .BackgroundMask
			oPageLayoutMap.ResizeMode = .ResizeMode
			oPageLayoutMap.DragMode = .DragMode
			oMapPrinterLayer.Layers.Clear()
		End With
	End Sub
	Public Sub AddMagneticNorthArrowElement()
		Dim oPrinterOverlay = mGetPrinterOverlay()
		Dim sKey As String
		Dim oMagDecLayer As CustomMagDecNorthArrow

		sKey = mGetKeyForMapElement(csMagneticNorthArrowBaseKey, moActiveLayoutMap)
		oMagDecLayer = New CustomMagDecNorthArrow(moActiveLayoutMap, sKey)

		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

		oPrinterOverlay.PrinterLayers.Add(sKey, oMagDecLayer)
		mSetPosition(1, 1, oMagDecLayer, EnumPagePosition.LowerLeft)


	End Sub
	Public Sub AddPageLayoutMap(oMap As CustomWinformsMap)
		Dim oPageBoundingbox As RectangleShape = GetPageBoundingBox(PrintingUnit.Inch)
		Dim oPageLayoutMap As CustomMapPrinterLayer = Nothing
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim bFoundMatch As Boolean = False
		Dim sKey As String

		'***** Note: This routine is called when creating a new page layout map from scratch
		'***** and also when opening a saved working set. If a working set is being opened 
		'***** the page layout map that goes with oMap will already exist, in which case
		'***** it just needs to be paired with oMap, the parent map. Otherwise a new page
		'***** layout map must be created and paired with its parent map, oMap.

		'***** See if there is already a CustomPageLayoutMap for this oMap
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = GetPageLayoutMaps()
		If oPageLayoutMaps Is Nothing Then
			'***** We are adding the first page layout map to the
			'***** PageLayout, so reset the counter
			miMapNumber = 0
		Else
			For Each oPageLayoutMap In oPageLayoutMaps
				If oPageLayoutMap.PageLayoutMapNumber = oMap.PageLayoutMapNumber Then
					'oPageLayoutMap.ParentMap = oMap
					bFoundMatch = True
					miMapNumber = oMap.PageLayoutMapNumber
					Exit For
				End If
			Next oPageLayoutMap
		End If

		If Not bFoundMatch Then
			oPageLayoutMap = New CustomMapPrinterLayer
			oPageLayoutMap.ResizeMode = PrinterResizeMode.MaintainAspectRatio

			'***** Create the MapPrinterLayer and set the position
			oPageLayoutMap.BackgroundMask = DisplayUtilities.CreateSimpleAreaStyle(GeoColor.StandardColors.White, GeoColor.SimpleColors.Black, 1)

			'***** Set the maps position at the page's center 7.5 inches wide and 10 inches tall
			oPageBoundingbox = GetPageBoundingBox(PrintingUnit.Inch)
			If miMapNumber = 0 Then
				oPageLayoutMap.SetPosition(7.5, 10, oPageBoundingbox.GetCenterPoint().X, oPageBoundingbox.GetCenterPoint().Y, PrintingUnit.Inch)
			Else
				oPageLayoutMap.SetPosition(4.75, 3, 0, 0, PrintingUnit.Inch)
			End If

			miMapNumber += 1
			oMap.PageLayoutMapNumber = miMapNumber
			oPageLayoutMap.PageLayoutMapNumber = miMapNumber

			'***** Get a key for adding the page layout map
			sKey = mGetKeyForMapElement(csMapPrinterLayerBaseKey, oPageLayoutMap)
			'***** Make the key the name of the page layout map
			oPageLayoutMap.Name = sKey
			'***** Add the MapPrinterLayer to the PrinterInteractiveOverlay
			oPrinterOverlay.PrinterLayers.Add(sKey, oPageLayoutMap)
		End If

		'***** Make this page layout map the 'Active' one
		UnSelectAllPageLayoutMaps()
		oPageLayoutMap.SelectMap()
		ActiveLayoutMap = oPageLayoutMap
		oPageLayoutMap.ParentMap = oMap

		'***** Set the MapUnit 
		ActiveMapUnit = oMap.MapUnit
	End Sub
	Public Sub RemovePageLayoutMap(iMapNumber As Integer)
		'***** Remove this layout map (MapPrinterLayer)
		'***** and all the surrounds that go with it
		Dim sKey As String = csMapPrinterLayerBaseKey & "_" & iMapNumber.ToString
		Dim oPrinterOverlay = mGetPrinterOverlay()
		oPrinterOverlay.PrinterLayers.Remove(sKey)

		sKey = csLegendLayerBaseKey & "_" & iMapNumber.ToString
		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

		sKey = csScaleBarLayerBaseKey & "_" & iMapNumber.ToString
		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

		sKey = csScaleLineLayerBaseKey & "_" & iMapNumber.ToString
		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

	End Sub
	Public Sub UnSelectAllPageLayoutMaps()
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = GetPageLayoutMaps()
		If Not oPageLayoutMaps Is Nothing Then
			For Each oPageLayoutMap As CustomMapPrinterLayer In oPageLayoutMaps
				oPageLayoutMap.UnSelectMap()
			Next oPageLayoutMap
		End If

	End Sub


	Private Function mGetMapRectPdfCoords(oMapRect As RectangleShape, eOrientation As PrinterOrientation) As RectangleShape
		Dim oPageRect As RectangleShape = GetPageBoundingBox(PrintingUnit.Inch)
		Dim dPageWidth As Double = oPageRect.Width * 72
		Dim dPageHeight As Double = oPageRect.Height * 72

		Dim dXMin As Double = dPageWidth / 2 + oMapRect.UpperLeftPoint.X * 72
		Dim dXMax As Double = dPageWidth / 2 + oMapRect.UpperRightPoint.X * 72
		Dim dYMin As Double = dPageHeight / 2 + oMapRect.LowerLeftPoint.Y * 72
		Dim dYMax As Double = dPageHeight / 2 + oMapRect.UpperLeftPoint.Y * 72

		'If eOrientation = PrinterOrientation.Portrait Then
		Return New RectangleShape(dXMin, dYMax, dXMax, dYMin)
		'Else
		'   '***** This must be done so that the geo-enabled pdf displays properly
		'   '***** when creating a pdf in landscape
		'   Return New RectangleShape(dXMin, dXMax, dYMax, dYMin)
		'End If

	End Function
	Private Function mGetPrinterCanvasDrawingArea(oPrintDocument As PrintDocument) As Rectangle
		Dim oRectangle As Rectangle
		Dim iOriginX As Integer
		Dim iOriginY As Integer

		'***** Here we create a System.Drawing.Rectangle to be used as the drawing
		'***** area for the PrinterGeoCanvas that will draw the map on the printer.
		'***** Notice that when the page has been set to Landscape the Width and Height
		'***** must be reversed.

		'***** Note that the dimenstions of the paper size are used and not the PrintableArea
		'***** of the printer. See more comments in the Print routine for why this is so.
		With oPrintDocument.DefaultPageSettings
			If oPrintDocument.DefaultPageSettings.Landscape Then
				iOriginX = 0
				iOriginY = 0
				oRectangle = New Rectangle(iOriginX, iOriginY, Convert.ToInt32(.PaperSize.Height), Convert.ToInt32(.PaperSize.Width))
			Else
				iOriginX = 0
				iOriginY = 0
				oRectangle = New Rectangle(iOriginX, iOriginY, Convert.ToInt32(.PaperSize.Width), Convert.ToInt32(.PaperSize.Height))
			End If
		End With

		Return oRectangle
	End Function
	Public Sub AddDataGridElement()
		Dim oPageLayoutMap As CustomMapPrinterLayer = ActiveLayoutMap
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim sKey As String = mGetKeyForMapElement(csDataGridBaseKey, moActiveLayoutMap)
		Dim oSqlLayer As CustomSQLFeatureLayer = Nothing
		Dim oMap As CustomWinformsMap = CurrentParentMap
		Dim oSqlConnectionInfo As SQLServerConnectionInfo = Nothing
		Dim oDataGridLayer As New CustomDataGridPrinterLayer
		Dim dHeight As Double

		With oDataGridLayer
			.BackgroundMask = DisplayUtilities.CreateSimpleAreaStyle(GeoColor.SimpleColors.White, GeoColor.SimpleColors.Black, 1)
			.BackgroundMask.DrawingLevel = DrawingLevel.LabelLevel
			.DragMode = PrinterDragMode.Draggable
			.ResizeMode = PrinterResizeMode.Resizable
		End With

		'***** Use the primary database if available, otherwise use the default
		'***** connection from preferences
		If Utils.CurrentPlugin.PrimaryDatabaseConnectionInfo Is Nothing Then
			Dim oConnectionInfoCollection = New SQLConnectionCollection
			oConnectionInfoCollection = DatabaseUtils.GetSqlConnectionCollection()
			If oConnectionInfoCollection.Contains(Utils.CurrentPlugin.Preferences.DefaultDatabaseConnection) Then
				oSqlConnectionInfo = oConnectionInfoCollection(Utils.CurrentPlugin.Preferences.DefaultDatabaseConnection)
			End If
		Else
			oSqlConnectionInfo = Utils.CurrentPlugin.PrimaryDatabaseConnectionInfo
		End If

		'***** It is possible that there is no Primary database connection and also no 
		'***** Default datbase connection. In that case the user will need to select/create
		'***** a database connection, without one the query dialog is useless.
		If oSqlConnectionInfo Is Nothing Then
			If Not DatabaseUtils.GetSqlConnectionInfo(oSqlConnectionInfo, ChooseConnectionCaption, Nothing, Utils.CurrentPlugin.ID, Utils.CurrentPlugin.Preferences.DefaultDatabaseConnection, False) Then
				MessageBox.Show(MessageConnectionRequiredForDataGrid, Utils.DefaultMessageCaption, MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
				Return
			End If
		End If


		Using oGridForm As New frmEditDataGrid(oDataGridLayer, oSqlConnectionInfo, oPageLayoutMap)
			oGridForm.ShowDialog()
			If Not oGridForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
				oDataGridLayer.MapPrinterLayer = oPageLayoutMap
				If Not oPageLayoutMap Is Nothing Then
					With oDataGridLayer
						.Name = sKey
						.ResizeMode = PrinterResizeMode.Resizable
						.TextFont = New GeoFont("Calibri", 8)
						.BackgroundMask.FillSolidBrush.Color = GeoColor.SimpleColors.White
						.HeaderTextFont = New GeoFont("Calibri", 8, DrawingFontStyles.Bold)
						.FillDataTable()
						.DataGridResizeMode = DataGridResizeMode.AutoFit
						Using oGraphics As System.Drawing.Graphics = Me.CreateGraphics
							dHeight = .GetGridHeight(oGraphics)
						End Using
						.Open()
						mSetPosition(4, dHeight, oDataGridLayer, EnumPagePosition.TopLeft)
						'***** When using DataGridResizeMode = DataGridResizeMode.AutoFit, simulating a MouseUp
						'***** followed by a MouseDown causes the data grid to do an 'Auto Resize' that corrects
						'***** any errors in its width. Don't know why it works, but it does
						'***** SET 01/27/2016
						mSimulateMouseUpDown()
					End With
				End If
				oPrinterOverlay.PrinterLayers.Add(sKey, oDataGridLayer)
			End If
		End Using
	End Sub
	Public Sub AddScaleText(oParentMap As CustomWinformsMap)
		Dim oScaletextPrinterLayer As ScaleTextPrinterLayer
		Dim dWidthInches As Double
		Dim dHeightInches As Double
		Dim oSizeF As SizeF
		Dim oGeoFont As GeoFont
		Dim sKey = mGetKeyForMapElement(csScaleTextLayerBaseKey, ActiveLayoutMap)

		moGraphics.PageUnit = GraphicsUnit.Pixel

		oGeoFont = New GeoFont("Arial", 10)
		oScaletextPrinterLayer = New ScaleTextPrinterLayer(ActiveLayoutMap, oGeoFont, New GeoSolidBrush(GeoColor.SimpleColors.Black))
		oScaletextPrinterLayer.Name = sKey
		With oScaletextPrinterLayer
			oSizeF = moGraphics.MeasureString(.Text, DisplayUtilities.FontFromGeoFont(oGeoFont))
			dHeightInches = 0.25
			dWidthInches = dHeightInches * (oSizeF.Width / oSizeF.Height)
			mSetPosition(dWidthInches, dHeightInches, oScaletextPrinterLayer, EnumPagePosition.TopLeft)
		End With

		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

		oPrinterOverlay.PrinterLayers.Add(sKey, oScaletextPrinterLayer)

	End Sub

	Public Sub CreateDefaultScaleLine(oMap As CustomWinformsMap)
		Dim oPageLayoutMap As MapPrinterLayer = ActiveLayoutMap
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim sKey As String = mGetKeyForMapElement(csScaleLineLayerBaseKey, moActiveLayoutMap)

		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

		If Not oPageLayoutMap Is Nothing Then
			'***** Create the scaleLinePrinterLayer and pass in the MapPrinterLayer
			Dim oScaleLineLayer As New ScaleLinePrinterLayer(oPageLayoutMap)
			With oScaleLineLayer
				.Name = sKey
				.MapUnit = oPageLayoutMap.MapUnit
				.ResizeMode = PrinterResizeMode.MaintainAspectRatio
				.Open()
			End With

			mSetPosition(2.5, 0.25, oScaleLineLayer, EnumPagePosition.LowerLeft)

			'***** Add the ScaleLinePrinterLayer to the PrinterInteractiveOverlay
			oPrinterOverlay.PrinterLayers.Add(sKey, oScaleLineLayer)
		End If
	End Sub
	Public Sub CreateDefaultScaleBar()
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim sKey As String = mGetKeyForMapElement(csScaleBarLayerBaseKey, moActiveLayoutMap)

		If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
			oPrinterOverlay.PrinterLayers.Remove(sKey)
		End If

		'***** Create the scaleLinePrinterLayer and pass in the MapPrinterLayer
		Dim oScaleBarLayer As New ScaleBarPrinterLayer(moActiveLayoutMap)

		With oScaleBarLayer

			.Name = sKey
			.MapUnit = moActiveLayoutMap.MapUnit
			.AlternateBarBrush = New GeoSolidBrush(GeoColor.SimpleColors.Black)
			.TextStyle = New TextStyle("", New GeoFont("Arial", 8), New GeoSolidBrush(GeoColor.SimpleColors.Black))
			.BarBrush = New GeoSolidBrush(GeoColor.SimpleColors.BrightRed)
			.UnitFamily = UnitSystem.Imperial
			.MaxWidth = 700
			.MaskBrush = New GeoSolidBrush(GeoColor.SimpleColors.Blue)
			.Height = 10
		End With

		oScaleBarLayer.Open()
		mSetPosition(2.5, 0.375, oScaleBarLayer, EnumPagePosition.LowerRight)

		'***** Add the ScaleLinePrinterLayer to the PrinterInteractiveOverlay
		oPrinterOverlay.PrinterLayers.Add(sKey, oScaleBarLayer)
	End Sub
	Public Sub BringElementToFront()
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

		If Not moSelectedElements Is Nothing Then
			If moSelectedElements.Count > 0 Then
				oPrinterOverlay.PrinterLayers.MoveToTop(moSelectedElements(0))
				moSelectedElements.MoveToTop(0)
				Refresh()
			End If
		End If
	End Sub
	Public Sub SendElementToBack()
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

		If Not moSelectedElements Is Nothing Then
			If moSelectedElements.Count > 0 Then
				oPrinterOverlay.PrinterLayers.MoveTo(moSelectedElements(0), 1)
				moSelectedElements.MoveToBottom(0)
				Refresh()
			End If
		End If
	End Sub

	Private Function mGetPrintPreviewPaperSize(ByVal oPagePrinterLayer As CustomPagePrinterLayer, sPrinterName As String) As PaperSize
		Dim oPaperSize As PaperSize = New PaperSize("AnsiA", 850, 1100)
		Dim bAdjustForLandscapePDF As Boolean

		'***** If printing to a pdf, let's not use landscape, just reverse the dimensions of
		'***** the page and print portrait. This makes it easier to create georeferenced PDFs.
		bAdjustForLandscapePDF = (oPagePrinterLayer.Orientation = PrinterOrientation.Landscape And sPrinterName.Contains("PDF"))

		Select Case oPagePrinterLayer.PageSize
			Case PrinterPageSize.AnsiA
				If bAdjustForLandscapePDF Then
					oPaperSize = New PaperSize("Custom", 1100, 850)
				Else
					oPaperSize = New PaperSize("AnsiA", 850, 1100)
				End If
			Case PrinterPageSize.AnsiB
				If bAdjustForLandscapePDF Then
					oPaperSize = New PaperSize("Custom", 1700, 1100)
				Else
					oPaperSize = New PaperSize("AnsiB", 1100, 1700)
				End If
			Case PrinterPageSize.AnsiC
				If bAdjustForLandscapePDF Then
					oPaperSize = New PaperSize("Custom", 2200, 1700)
				Else
					oPaperSize = New PaperSize("AnsiC", 1700, 2200)
				End If
			Case PrinterPageSize.AnsiD
				If bAdjustForLandscapePDF Then
					oPaperSize = New PaperSize("Custom", 3400, 2200)
				Else
					oPaperSize = New PaperSize("AnsiD", 2200, 3400)
				End If
			Case PrinterPageSize.AnsiE
				If bAdjustForLandscapePDF Then
					oPaperSize = New PaperSize("Custom", 4400, 3400)
				Else
					oPaperSize = New PaperSize("AnsiE", 3400, 4400)
				End If
			Case PrinterPageSize.Custom
				oPaperSize = New PaperSize("Custom Size", CInt(oPagePrinterLayer.CustomWidth), CInt(oPagePrinterLayer.CustomHeight))
			Case Else
		End Select

		Return oPaperSize
	End Function

	Private Function GetPageBoundingBox(ByVal unit As PrintingUnit) As RectangleShape
		'***** This helper method gets the pages bounding box in the unit requested
		Dim oPrinterPage As CustomPagePrinterLayer = mGetPrinterPage()
		Return oPrinterPage.GetPosition(unit)
	End Function
	Public Sub SetGeoExtent(oMap As CustomWinformsMap, ByVal oExtent As RectangleShape)
		ActiveLayoutMap.MapExtent = oExtent
	End Sub
	Public Sub ViewEntirePageLayout(Optional eRefreshBehavior As Enums.EnumRefreshBehavior = Enums.EnumRefreshBehavior.Refresh)
		If Not PrinterPage Is Nothing Then
			If Not PrinterPage.IsOpen Then
				PrinterPage.Open()
			End If
			Dim oCurrentExtent As RectangleShape = PrinterPage.GetBoundingBox

			'''''oCurrentExtent.ScaleUp(10)
			CurrentExtent = oCurrentExtent
			If eRefreshBehavior = Enums.EnumRefreshBehavior.Refresh Then
				Refresh(mGetPrinterOverlay)
			End If
		End If
	End Sub
	Public Property DrawPageLayoutBorder() As Boolean
		Get
			Return mGetPrinterPage.DrawPageBorder
		End Get
		Set(value As Boolean)
			mGetPrinterPage.DrawPageBorder = value
			Refresh()
		End Set
	End Property
	Public Sub EditPrinterPageBorderProps()
		Dim eResult As DialogResult
		Using oForm As New frmPageLayoutBorderProps(mGetPrinterPage)
			With oForm
				eResult = .ShowDialog()
				If eResult <> Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End With
		End Using
	End Sub
	Public ReadOnly Property PrinterPage As CustomPagePrinterLayer
		Get
			Return mGetPrinterPage()
		End Get

	End Property

	Public ReadOnly Property PrinterOverlay As PrinterInteractiveOverLay
		Get
			Return mGetPrinterOverlay()
		End Get
	End Property
	Private Function mGetPrinterOverlay() As PrinterInteractiveOverLay
		If moCurPrinterOverlay Is Nothing Then
			Dim sKey As String = CType(PageLayoutControl.cboPageNumber.SelectedItem, ListBoxItem).ValueMember.ToString
			Return moPrinterOverlays(sKey)
		Else
			Return moCurPrinterOverlay
		End If
	End Function
	Private Function mGetAutoLayoutFeature() As Feature
		Dim oCustomPrinterOverlay As CustomPrinterOverlay = TryCast(mGetPrinterOverlay(), CustomPrinterOverlay)
		Dim oFeature As Feature = Nothing

		If oCustomPrinterOverlay IsNot Nothing Then
			oFeature = oCustomPrinterOverlay.AutoLayoutFeature
		End If

		Return oFeature
	End Function
	Private Sub InitializeComponent()
		Me.SuspendLayout()
		'
		'SilvicsPageLayout
		'
		Me.ResumeLayout(False)

	End Sub

	Public Sub AddImageElement()
		Dim oDialog As New OpenFileDialog()

		'***** Create a new image layer and add it to the page layout
		Dim sFilter As String = "All image files|*.tif;*.tiff;*.jpg;*.jpeg;*.jp2;*.sid;*.png;*.ecw;*.bmp"
		Dim sFilename As String = String.Empty
		Dim dWidth As Double
		Dim dHeight As Double
		Dim eDialogResult As DialogResult

		With oDialog
			.Filter = sFilter
			.Title = "Choose file for image"
			.ShowReadOnly = False
			eDialogResult = .ShowDialog
			sFilename = .FileName
		End With

		If Not eDialogResult = Windows.Forms.DialogResult.Cancel Then
			'***** Add the image to the map
			Dim oGeoImage As New GeoImage(sFilename)
			'***** Create a width and height in inches in order to place the layer
			'***** on the map
			dWidth = 1.5
			dHeight = dWidth * (oGeoImage.Height / oGeoImage.Width)

			Dim oImagePrinterLayer As New ImagePrinterLayer(oGeoImage)
			Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

			oPrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oImagePrinterLayer)
			mSetPosition(dWidth, dHeight, oImagePrinterLayer, EnumPagePosition.TopLeft)
		End If

	End Sub
	Public Sub AddLabelElement()
		Dim oLabelPrinterLayer As LabelPrinterLayer
		Dim dWidthInches As Double
		Dim dHeightInches As Double
		Dim oSizeF As SizeF
		Dim oGeoFont As GeoFont

		moGraphics.PageUnit = GraphicsUnit.Pixel

		oGeoFont = New GeoFont("Arial", 10)
		oLabelPrinterLayer = New LabelPrinterLayer("New Label", oGeoFont, New GeoSolidBrush(GeoColor.SimpleColors.Black))
		With oLabelPrinterLayer
			.PrinterWrapMode = PrinterWrapMode.AutoSizeText
			.ResizeMode = PrinterResizeMode.MaintainAspectRatio
			oSizeF = moGraphics.MeasureString("New Label", DisplayUtilities.FontFromGeoFont(oGeoFont))
			dHeightInches = 0.25
			dWidthInches = dHeightInches * (oSizeF.Width / oSizeF.Height)
			mSetPosition(dWidthInches, dHeightInches, oLabelPrinterLayer, EnumPagePosition.TopLeft)
			'moPrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oLabelPrinterLayer)
		End With

		Dim oLabelForm As New frmEditPrinterLabel(oLabelPrinterLayer)
		oLabelForm.ShowDialog()
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

		If Not oLabelForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
			oPrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oLabelPrinterLayer)
		End If

		oLabelForm.Dispose()


	End Sub
	Public Sub mShowEditElementForm(oMap As CustomWinformsMap, oPrinterLayer As Object)
		Dim oPageLayoutMap As CustomMapPrinterLayer = Nothing
		If TypeOf oPrinterLayer Is LegendPrinterLayer Then
			Dim oLegendMap As CustomMapPrinterLayer = CType(oPrinterLayer, CustomLegendLayer).PageLayoutMap
			Using oForm As New frmEditLegend(oLegendMap)
				oForm.ShowDialog()
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End Using
		ElseIf TypeOf oPrinterLayer Is ScaleLinePrinterLayer Then
			Using oForm As New frmEditGenericePrinterLayer(CType(oPrinterLayer, PrinterLayer))
				oForm.ShowDialog()
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End Using
		ElseIf TypeOf oPrinterLayer Is ImagePrinterLayer Or TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
			If TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
				oPageLayoutMap = CType(oPrinterLayer, CustomMapPrinterLayer)
				oPageLayoutMap.BackgroundMask = oPageLayoutMap.UnSelectedBackGroundMask
			End If
			Using oForm As New frmEditGenericePrinterLayer(CType(oPrinterLayer, PrinterLayer))
				oForm.ShowDialog()
				If Not oPageLayoutMap Is Nothing Then
					oPageLayoutMap.UnSelectedBackGroundMask = CType(oPageLayoutMap.BackgroundMask.CloneDeep, AreaStyle)
					If oPageLayoutMap.IsSelected Then
						oPageLayoutMap.BackgroundMask = CType(oPageLayoutMap.SelectedBackgroundMask.CloneDeep, AreaStyle)
					End If
				End If
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End Using
		ElseIf TypeOf oPrinterLayer Is LabelPrinterLayer Then
			Using oForm As New frmEditPrinterLabel(CType(oPrinterLayer, LabelPrinterLayer))
				oForm.ShowDialog()
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End Using
		ElseIf TypeOf oPrinterLayer Is ScaleBarPrinterLayer Then
			Using oForm As New frmEditScaleBarLayer(CType(oPrinterLayer, ScaleBarPrinterLayer))
				oForm.ShowDialog()
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End Using
		ElseIf TypeOf oPrinterLayer Is MagneticDeclinationPrinterLayer Then
			Using oForm As New frmEditMagDecNorthArrow(CType(oPrinterLayer, CustomMagDecNorthArrow))
				oForm.ShowDialog()
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					Refresh()
				End If
			End Using
		ElseIf TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
			Using oForm As New frmEditDataGrid(CType(oPrinterLayer, CustomDataGridPrinterLayer), CType(oPrinterLayer, CustomDataGridPrinterLayer).ConnectionInfo, ActiveLayoutMap)
				oForm.ShowDialog()
				If Not oForm.DialogResult = Windows.Forms.DialogResult.Cancel Then
					mRefreshDataGrid(CType(oPrinterLayer, CustomDataGridPrinterLayer))
					Refresh()
				End If
			End Using

		End If

	End Sub
	Public ReadOnly Property PageLayoutControl As PageLayoutControl
		Get
			Dim oPageLayoutControl As PageLayoutControl = TryCast(Me.Parent, PageLayoutControl)
			Return oPageLayoutControl
		End Get
	End Property
	Private Sub PageLayoutControl_MapClick(ByVal sender As Object, ByVal e As ThinkGeo.MapSuite.WinForms.MapClickWinformsMapEventArgs) Handles Me.MapClick
		'***** See what we clicked on
		Dim oPrinterOverlay As PrinterInteractiveOverLay = PrinterOverlay
		Dim oPrinterLayer As PrinterLayer
		Dim oPrinterLayers As New Collection(Of PrinterLayer)
		moSelectedPrinterLayer = Nothing

		For Each oPrinterLayer In oPrinterOverlay.PrinterLayers
			If Not TypeOf oPrinterLayer Is CustomPagePrinterLayer And Not TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
				oPrinterLayer.Open()
				Dim oBoundingBox As RectangleShape = oPrinterLayer.GetPosition
				If oBoundingBox.Contains(e.WorldLocation) Then
					oPrinterLayers.Add(oPrinterLayer)
				End If
			End If
		Next oPrinterLayer

		If oPrinterLayers.Count > 0 Then
			moSelectedPrinterLayer = mGetLeastAreaPrinterLayer(oPrinterLayers)
		End If


		If moSelectedPrinterLayer Is Nothing Then
			For i As Integer = oPrinterOverlay.PrinterLayers.Count - 1 To 0 Step -1
				oPrinterLayer = oPrinterOverlay.PrinterLayers(i)
				'***** We didn't find a PrinterLayer that's not a MapPrinterLayer so let's
				'***** look for one that is
				If TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
					oPrinterLayer.Open()
					Dim oBoundingBox As RectangleShape = oPrinterLayer.GetPosition
					If oBoundingBox.Contains(e.WorldLocation) Then
						If e.MouseButton = MapMouseButton.Left Then
							SetCurrentPageLayout(CType(oPrinterLayer, CustomMapPrinterLayer))
						End If
						If e.MouseButton = MapMouseButton.Right Then
							moSelectedPrinterLayer = oPrinterLayer
						End If
						Exit For
					End If
				End If
			Next i
		End If

		If e.MouseButton = MapMouseButton.Right Then
			PageLayoutControl.ShowPopupMenu(moSelectedPrinterLayer)
		End If

	End Sub
	Private Function mGetLeastAreaPrinterLayer(oPrinterLayers As Collection(Of PrinterLayer)) As PrinterLayer
		Dim dMinArea As Double = Double.MaxValue
		Dim dPrinterLayerArea As Double
		Dim oLeastAreaPrinterLayer As PrinterLayer = Nothing

		For Each oPrinterLayer In oPrinterLayers
			dPrinterLayerArea = oPrinterLayer.GetBoundingBox.GetArea(GeographyUnit.Feet, AreaUnit.SquareFeet)
			If dPrinterLayerArea < dMinArea Then
				dMinArea = dPrinterLayerArea
				oLeastAreaPrinterLayer = oPrinterLayer
			End If
		Next oPrinterLayer

		Return oLeastAreaPrinterLayer
	End Function
	Public Sub SetCurrentPageLayout(Optional ByVal oPageLayoutMap As CustomMapPrinterLayer = Nothing, Optional eRefresh As Enums.EnumRefreshBehavior = Enums.EnumRefreshBehavior.Refresh)
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = GetPageLayoutMaps()
		Dim oPLMap As CustomMapPrinterLayer

		UnSelectAllPageLayoutMaps()

		If oPageLayoutMap Is Nothing Then
			If Not oPageLayoutMaps Is Nothing Then
				For Each oPLMap In oPageLayoutMaps
					If oPLMap.ParentMap.IsActiveMap Then
						oPageLayoutMap = oPLMap
						Exit For
					End If
				Next oPLMap
			End If
		End If

		If Not oPageLayoutMap Is Nothing Then
			ActiveLayoutMap = oPageLayoutMap
			oPageLayoutMap.SelectMap()
		End If

		If eRefresh = Enums.EnumRefreshBehavior.Refresh Then
			Refresh()
		End If


	End Sub
	Public Sub DeleteElement(ByVal oPrinterLayer As PrinterLayer)
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

		If Not oPrinterLayer Is Nothing Then
			If Not TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
				If mAskDeleteOK(oPrinterLayer) Then
					oPrinterOverlay.PrinterLayers.Remove(oPrinterLayer)
					Refresh()
				End If
			End If
		End If
		moSelectedPrinterLayer = Nothing
	End Sub
	Private Sub moPrinterOverlay_MapKeyUp(sender As Object, e As ThinkGeo.MapSuite.WinForms.MapKeyUpInteractiveOverlayEventArgs) 'Handles moPrinterOverlay.MapKeyUp
		If e.InteractionArguments.Key = Keys.Delete.ToString Then
			DeleteElement(moSelectedPrinterLayer)
		End If
	End Sub
	Public Function GetPageLayoutMaps() As Collection(Of CustomMapPrinterLayer)
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = Nothing

		For Each oPrinterLayer As PrinterLayer In oPrinterOverlay.PrinterLayers
			If TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
				If oPageLayoutMaps Is Nothing Then
					oPageLayoutMaps = New Collection(Of CustomMapPrinterLayer)
				End If
				oPageLayoutMaps.Add(CType(oPrinterLayer, CustomMapPrinterLayer))
			End If
		Next oPrinterLayer
		Return oPageLayoutMaps
	End Function
	Public Sub RefreshLegendLayers(oLayersToInclude As Collection(Of Layer),
																 oCurLegend As CustomLegendLayer,
																 eRefreshBehavior As Enums.EnumRefreshBehavior)
		Dim oLegendItem As CustomLegendItem = Nothing
		Dim oValueItem As CustomValueItem
		Dim oTextStyle As TextStyle
		Dim sText As String
		Dim oStyle As Style
		Dim sLayerKey As String = String.Empty
		Dim sItemKey As String = String.Empty
		Dim oPageBoundingBox As RectangleShape
		Dim iValueItemCounter As Integer
		Dim oLayer As Layer
		Dim eShapeType As ShapeFileType
		Dim sLegendKey As String = mGetKeyForMapElement(csLegendLayerBaseKey, ActiveLayoutMap)
		Dim oPrinterPage As CustomPagePrinterLayer = PrinterPage

		oPrinterPage.Open()
		oPageBoundingBox = oPrinterPage.GetBoundingBox
		oCurLegend.LegendItems.Clear()

		'***** Set the title LegendItem for the legend
		oTextStyle = New TextStyle(oCurLegend.TitleText, oCurLegend.TitleFont, New GeoSolidBrush(oCurLegend.TitleFontColor))
		oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.Title)
		oCurLegend.Title = oLegendItem

		For Each oLayer In oLayersToInclude
			If TypeOf oLayer Is ICustomLayer Then
				If CType(oLayer, ICustomLayer).IsStealthEditingLayer Then
					Continue For
				End If
			End If
			sLayerKey = oLayer.Name
			'***** Get the style being used for this layer's symbology
			oStyle = LayerUtils.GetStyleForLayer(oLayer, GetScale())
			If TypeOf oStyle Is CustomValueStyle Then
				eShapeType = CType(oLayer, ICustomLayer).ShapeType
				iValueItemCounter = 0

				'***** Add a legend item for the name of the layer
				oTextStyle = New TextStyle(oLayer.Name, oCurLegend.LayerFont, New GeoSolidBrush(oCurLegend.LayerFontColor))
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.FeatureLayer)
				oLegendItem.Layer = oLayer
				oCurLegend.LegendItems.Add(sLayerKey, oLegendItem)

				'***** Add a legend item for the column that's being themed on
				oTextStyle = New TextStyle(CType(oStyle, CustomValueStyle).Title, oCurLegend.ItemFont, New GeoSolidBrush(oCurLegend.ItemFontColor))
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, ThemeHeader)
				iValueItemCounter += 1
				sItemKey = sLayerKey & " - " & iValueItemCounter
				oCurLegend.LegendItems.Add(sItemKey, oLegendItem)

				'***** Before adding the individual value items to the legend, 
				'***** add the 'Default' symbol
				If CType(oStyle, CustomValueStyle).UseDefaultSymbol Then
					oTextStyle = New TextStyle(CType(oStyle, CustomValueStyle).DefaultDisplayText, oCurLegend.DefaultLayerFont, New GeoSolidBrush(GeoColor.SimpleColors.Black))
					oLegendItem = mCreateDefaultLegendItem(CType(oStyle, CustomValueStyle).DefaultStyle, oTextStyle, ThemeItem)
					iValueItemCounter += 1
					sItemKey = sLayerKey & " - " & iValueItemCounter
					oCurLegend.LegendItems.Add(sItemKey, oLegendItem)
				End If

				For Each oValueItem In CType(oStyle, ValueStyle).ValueItems
					If oValueItem.UseValue Then
						If oValueItem.DisplayValue = String.Empty Then
							sText = "Null"
						Else
							sText = oValueItem.DisplayValue
						End If
						oTextStyle = New TextStyle(sText, oCurLegend.ItemFont, New GeoSolidBrush(oCurLegend.ItemFontColor))
						Select Case eShapeType
							Case ShapeFileType.Point
								oLegendItem = mCreateDefaultLegendItem(oValueItem.DefaultPointStyle, oTextStyle, ThemeItem)
							Case ShapeFileType.Polyline
								oLegendItem = mCreateDefaultLegendItem(oValueItem.DefaultLineStyle, oTextStyle, ThemeItem)
							Case ShapeFileType.Polygon
								oLegendItem = mCreateDefaultLegendItem(oValueItem.DefaultAreaStyle, oTextStyle, ThemeItem)
						End Select
						iValueItemCounter += 1
						sItemKey = sLayerKey & " - " & iValueItemCounter
						oCurLegend.LegendItems.Add(sItemKey, oLegendItem)
					End If
				Next oValueItem
			ElseIf TypeOf oStyle Is CustomClassBreakStyle Then
				iValueItemCounter = 0
				eShapeType = CType(oLayer, ICustomLayer).ShapeType
				'***** Add a legend item for the name of the layer
				oTextStyle = New TextStyle(oLayer.Name, oCurLegend.LayerFont, New GeoSolidBrush(oCurLegend.LayerFontColor))
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.FeatureLayer)
				oLegendItem.Layer = oLayer
				oCurLegend.LegendItems.Add(sLayerKey, oLegendItem)

				'***** Add a legend item for the column that's being themed on
				oTextStyle = New TextStyle(CType(oStyle, CustomClassBreakStyle).Title, oCurLegend.LayerFont, New GeoSolidBrush(oCurLegend.LayerFontColor))
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.ThemeHeader)
				iValueItemCounter += 1
				sItemKey = sLayerKey & " - " & iValueItemCounter
				oCurLegend.LegendItems.Add(sItemKey, oLegendItem)

				For Each oClassBreak As CustomClassBreak In CType(oStyle, CustomClassBreakStyle).ClassBreaks
					'If oValueItem.UseValue Then
					sText = oClassBreak.DisplayValue
					oTextStyle = New TextStyle(sText, oCurLegend.ItemFont, New GeoSolidBrush(oCurLegend.ItemFontColor))
					Select Case eShapeType
						Case ShapeFileType.Point
							oLegendItem = mCreateDefaultLegendItem(oClassBreak.DefaultPointStyle, oTextStyle, ThemeItem)
						Case ShapeFileType.Polyline
							oLegendItem = mCreateDefaultLegendItem(oClassBreak.DefaultLineStyle, oTextStyle, ThemeItem)
						Case ShapeFileType.Polygon
							oLegendItem = mCreateDefaultLegendItem(oClassBreak.DefaultAreaStyle, oTextStyle, ThemeItem)
					End Select
					iValueItemCounter += 1
					sItemKey = sLayerKey & " - " & iValueItemCounter
					oCurLegend.LegendItems.Add(sItemKey, oLegendItem)
					'End If
				Next oClassBreak

			Else
				'***** Just add the layer name with it's symbol
				oTextStyle = New TextStyle(oLayer.Name, oCurLegend.LayerFont, New GeoSolidBrush(oCurLegend.LayerFontColor))
				If TypeOf oLayer Is FeatureLayer Then
					oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.FeatureLayer)
				ElseIf TypeOf oLayer Is WmsRasterLayer Then
					oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.WMSLayer)
				ElseIf TypeOf oLayer Is RasterLayer Then
					oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.RasterLayer)
				ElseIf TypeOf oLayer Is OpenStreetMapLayer Then
					oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.OpenStreetMapsLayer)
				ElseIf TypeOf oLayer Is GoogleMapsLayer Then
					oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.GoogleMapsLayer)
				End If
				oCurLegend.LegendItems.Add(sLayerKey, oLegendItem)
				oLegendItem.Layer = oLayer
			End If
			sLayerKey = String.Empty
		Next oLayer


		Dim dWidthInPixels = mGetLegendWidth(oCurLegend)
		Dim dHeightInPixels = mGetLegendHeight(oCurLegend)

		Dim dWidthInInches As Double
		Dim dHeightInInches As Double

		dWidthInInches = dWidthInPixels / msngDpi
		dHeightInInches = dHeightInPixels / msngDpi

		'***** Do not change the actual width or height of the legend 
		'With oCurLegend
		'   .Width = dWidthInPixels
		'   .Height = CSng(dHeightInPixels * 0.9)
		'End With
		'***** Been going back and forth on this....
		'***** The legend is messed up when doing auto layouts if we don't resize it here
		'***** SET 04/07/2016
		With oCurLegend
			.Width = dWidthInPixels
			.Height = CSng(dHeightInPixels * 0.95)
			dWidthInInches = dWidthInPixels / msngDpi
			dHeightInInches = dHeightInPixels / msngDpi
		End With


		oCurLegend.ResizeMode = PrinterResizeMode.Resizable
		oCurLegend.Open()

		Dim oCurPosition = oCurLegend.GetPosition(PrintingUnit.Inch)
		oCurLegend.SetPosition(dWidthInInches, dHeightInInches, oCurPosition.GetCenterPoint, PrintingUnit.Inch)

		If eRefreshBehavior = Enums.EnumRefreshBehavior.Refresh Then
			Refresh()
		End If

	End Sub

	Public Sub BuildLegend(eBuildLegendOption As EnumBuildLegendOption)
		Dim oPrinterInteractiveOverlay As PrinterInteractiveOverLay
		Dim oPageLayoutMap As CustomMapPrinterLayer = ActiveLayoutMap
		Dim oNewLegendLayer As CustomLegendLayer
		Dim oLegendItem As CustomLegendItem
		Dim oValueItem As CustomValueItem
		Dim oTextStyle As TextStyle
		Dim sText As String
		Dim oStyle As Style
		Dim sLayerKey As String = String.Empty
		Dim sItemKey As String = String.Empty
		Dim oPageBoundingBox As RectangleShape = PrinterPage.GetBoundingBox
		Dim iValueItemCounter As Integer
		Dim oBlackBrush As New GeoSolidBrush(GeoColor.SimpleColors.Black)
		Dim oLayer As Layer
		Dim oTitleFont As GeoFont = Nothing
		Dim oItemFont As GeoFont = Nothing
		Dim oLayerFont As GeoFont
		Dim dWidthInInches As Double
		Dim dHeightInInches As Double
		Dim eShapeType As ShapeFileType
		Dim sLegendKey As String
		Dim oCurrentLegendLayer As CustomLegendLayer = Nothing

		Dim oExistingLegendItems As GeoCollection(Of LegendItem) = Nothing

		'***** See if there is a legend layer and if there is, get rid of it
		oPrinterInteractiveOverlay = mGetPrinterOverlay()
		sLegendKey = mGetKeyForMapElement(csLegendLayerBaseKey, moActiveLayoutMap)

		If oPrinterInteractiveOverlay.PrinterLayers.Contains(sLegendKey) Then
			oCurrentLegendLayer = mGetLegend(sLegendKey)
			oPrinterInteractiveOverlay.PrinterLayers.Remove(sLegendKey)
		End If

		'***** This is what we used to do, but now we have a clone method
		'***************************************************************************
		'''''''''''''''''oNewLegendLayer = New CustomLegendLayer(sLegendKey, Me)
		'***************************************************************************
		'oPageLayoutMap.Legend = oNewLegendLayer
		'oNewLegendLayer.PageLayoutMap = ActiveLayoutMap

		If eBuildLegendOption = EnumBuildLegendOption.CreateDefaultLegend Then
			'***** Create legend from scratch
			oNewLegendLayer = New CustomLegendLayer(sLegendKey, Me)
			oPageLayoutMap.Legend = oNewLegendLayer
			oNewLegendLayer.PageLayoutMap = ActiveLayoutMap


			'***** Get the fonts to be used for the legend items
			oTitleFont = oNewLegendLayer.DefaultTitleFont
			oItemFont = oNewLegendLayer.DefaultItemFont
			oLayerFont = oNewLegendLayer.DefaultLayerFont
			'***** Set the title LegendItem for the legend
			oTextStyle = New TextStyle("Legend", oTitleFont, oBlackBrush)
			oNewLegendLayer.TitleText = "Legend"
			oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, Title)
			oNewLegendLayer.Title = oLegendItem
			oNewLegendLayer.LegendItems.Clear()
		Else
			'***** Create legend by cloning existing legend
			oNewLegendLayer = oCurrentLegendLayer.Clone()
			oPageLayoutMap.Legend = oNewLegendLayer
			oNewLegendLayer.PageLayoutMap = ActiveLayoutMap
			oNewLegendLayer.SilvicsPageLayout = Me

			'***** Get the fonts for the title, layer, and theme items
			oTitleFont = oCurrentLegendLayer.Title.TextStyle.Font

			oLegendItem = oCurrentLegendLayer.GetFirstLegendItem(CustomLegendItem.EnumLegendItemType.FeatureLayer)
			oLayerFont = oLegendItem.TextStyle.Font

			oLegendItem = oCurrentLegendLayer.GetFirstLegendItem(ThemeItem)
			If oLegendItem IsNot Nothing Then
				oItemFont = oLegendItem.TextStyle.Font
			End If

			''''oNewLegendLayer.Title = oCurrentLegendLayer.Title
			''''oNewLegendLayer.Title.TextStyle.Font = oTitleFont
			oExistingLegendItems = New GeoCollection(Of LegendItem)
			For Each oItem In oNewLegendLayer.LegendItems
				oExistingLegendItems.Add(oItem)
			Next
			oNewLegendLayer.LegendItems.Clear()
		End If

		For i As Integer = oPageLayoutMap.Layers.Count - 1 To 0 Step -1
			oLayer = oPageLayoutMap.Layers(i)
			If TypeOf oLayer Is ICustomLayer Then
				If CType(oLayer, ICustomLayer).IsStealthEditingLayer Then
					Continue For
				End If
			End If
			If eBuildLegendOption = EnumBuildLegendOption.CopyExistingLegend Then
				If Not CType(oLayer, ILegendProperties).IncludeInLegend Then
					Continue For
				End If
			Else
				CType(oLayer, ILegendProperties).IncludeInLegend = True
			End If
			sLayerKey = oLayer.Name
			'***** Get the style being used for this layer's symbology
			oStyle = LayerUtils.GetStyleForLayer(oLayer, GetScale())
			If TypeOf oStyle Is ValueStyle Then
				iValueItemCounter = 0
				eShapeType = CType(oLayer, ICustomLayer).ShapeType
				'***** Add a legend item for the name of the layer
				oTextStyle = New TextStyle(oLayer.Name, oLayerFont, oBlackBrush)
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.FeatureLayer)
				oLegendItem.Layer = oLayer
				oNewLegendLayer.LegendItems.Add(sLayerKey, oLegendItem)

				'***** Add a legend item for the column that's being themed on
				oTextStyle = New TextStyle(CType(oStyle, CustomValueStyle).Title, oLayerFont, oBlackBrush)
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.ThemeHeader)
				iValueItemCounter += 1
				sItemKey = sLayerKey & " - " & iValueItemCounter
				oNewLegendLayer.LegendItems.Add(sItemKey, oLegendItem)

				'***** Before adding the individual value items to the legend, 
				'***** add the 'Default' symbol
				If CType(oStyle, CustomValueStyle).UseDefaultSymbol Then
					oTextStyle = New TextStyle(CType(oStyle, CustomValueStyle).DefaultDisplayText, oItemFont, oBlackBrush)
					oLegendItem = mCreateDefaultLegendItem(CType(oStyle, CustomValueStyle).DefaultStyle, oTextStyle, ThemeItem)
					iValueItemCounter += 1
					sItemKey = sLayerKey & " - " & iValueItemCounter
					oNewLegendLayer.LegendItems.Add(sItemKey, oLegendItem)
				End If

				For Each oValueItem In CType(oStyle, CustomValueStyle).ValueItems
					If oValueItem.UseValue Then
						If oValueItem.DisplayValue = String.Empty Then
							sText = "Null"
						Else
							sText = oValueItem.DisplayValue
						End If
						oTextStyle = New TextStyle(sText, oItemFont, oBlackBrush)
						Select Case eShapeType
							Case ShapeFileType.Point
								oLegendItem = mCreateDefaultLegendItem(oValueItem.DefaultPointStyle, oTextStyle, ThemeItem)
							Case ShapeFileType.Polyline
								oLegendItem = mCreateDefaultLegendItem(oValueItem.DefaultLineStyle, oTextStyle, ThemeItem)
							Case ShapeFileType.Polygon
								oLegendItem = mCreateDefaultLegendItem(oValueItem.DefaultAreaStyle, oTextStyle, ThemeItem)
						End Select
						iValueItemCounter += 1
						sItemKey = sLayerKey & " - " & iValueItemCounter
						oNewLegendLayer.LegendItems.Add(sItemKey, oLegendItem)
					End If
				Next oValueItem
			ElseIf TypeOf oStyle Is CustomClassBreakStyle Then
				iValueItemCounter = 0
				eShapeType = CType(oLayer, ICustomLayer).ShapeType
				'***** Add a legend item for the name of the layer
				oTextStyle = New TextStyle(oLayer.Name, oLayerFont, oBlackBrush)
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.FeatureLayer)
				oLegendItem.Layer = oLayer
				oNewLegendLayer.LegendItems.Add(sLayerKey, oLegendItem)

				'***** Add a legend item for the column that's being themed on
				oTextStyle = New TextStyle(CType(oStyle, CustomClassBreakStyle).Title, oLayerFont, oBlackBrush)
				oLegendItem = mCreateDefaultLegendItem(Nothing, oTextStyle, CustomLegendItem.EnumLegendItemType.ThemeHeader)
				iValueItemCounter += 1
				sItemKey = sLayerKey & " - " & iValueItemCounter
				oNewLegendLayer.LegendItems.Add(sItemKey, oLegendItem)

				For Each oClassBreak As CustomClassBreak In CType(oStyle, CustomClassBreakStyle).ClassBreaks
					'If oValueItem.UseValue Then
					sText = oClassBreak.DisplayValue
					oTextStyle = New TextStyle(sText, oItemFont, oBlackBrush)
					Select Case eShapeType
						Case ShapeFileType.Point
							oLegendItem = mCreateDefaultLegendItem(oClassBreak.DefaultPointStyle, oTextStyle, ThemeItem)
						Case ShapeFileType.Polyline
							oLegendItem = mCreateDefaultLegendItem(oClassBreak.DefaultLineStyle, oTextStyle, ThemeItem)
						Case ShapeFileType.Polygon
							oLegendItem = mCreateDefaultLegendItem(oClassBreak.DefaultAreaStyle, oTextStyle, ThemeItem)
					End Select
					iValueItemCounter += 1
					sItemKey = sLayerKey & " - " & iValueItemCounter
					oNewLegendLayer.LegendItems.Add(sItemKey, oLegendItem)
					'End If
				Next oClassBreak

			Else
				'***** Just add the layer name with it's symbol
				If CType(oLayer, ILegendProperties).IncludeInLegend Then
					oTextStyle = New TextStyle(oLayer.Name, oLayerFont, oBlackBrush)
					If TypeOf oLayer Is FeatureLayer Then
						oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.FeatureLayer)
					ElseIf TypeOf oLayer Is WmsRasterLayer Then
						oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.WMSLayer)
					ElseIf TypeOf oLayer Is RasterLayer Then
						oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.RasterLayer)
					ElseIf TypeOf oLayer Is OpenStreetMapLayer Then
						oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.OpenStreetMapsLayer)
					ElseIf TypeOf oLayer Is GoogleMapsLayer Then
						oLegendItem = mCreateDefaultLegendItem(oStyle, oTextStyle, CustomLegendItem.EnumLegendItemType.GoogleMapsLayer)
					End If
				End If
				oLegendItem.Layer = oLayer
				oNewLegendLayer.LegendItems.Add(sLayerKey, oLegendItem)
			End If
			sLayerKey = String.Empty
			oStyle = Nothing
		Next i

		Dim dWidthInPixels = mGetLegendWidth(oNewLegendLayer)
		Dim dHeightInPixels = mGetLegendHeight(oNewLegendLayer)

		With oNewLegendLayer
			.Width = dWidthInPixels
			.Height = CSng(dHeightInPixels * 0.95)
			dWidthInInches = dWidthInPixels / msngDpi
			dHeightInInches = dHeightInPixels / msngDpi
		End With
		oNewLegendLayer.ResizeMode = PrinterResizeMode.Resizable
		oPrinterInteractiveOverlay.PrinterLayers.Add(sLegendKey, oNewLegendLayer)
		oNewLegendLayer.Open()

		If eBuildLegendOption = EnumBuildLegendOption.CreateDefaultLegend Then
			mSetPosition(dWidthInInches, dHeightInInches, oNewLegendLayer, EnumPagePosition.LowerRight)
		Else
			Dim oCurPosition = oCurrentLegendLayer.GetPosition(PrintingUnit.Inch)
			'oNewLegendLayer.SetPosition(dWidthInInches, dHeightInInches, oCurPosition.GetCenterPoint.X / msngDpi, oCurPosition.GetCenterPoint.Y / msngDpi, PrintingUnit.Inch)
			'***** The method of positioning an existing legend below preserves its spot in the page.
			'***** SET 10/20/2017
			oNewLegendLayer.SetPosition(dWidthInInches, dHeightInInches, oCurPosition.GetCenterPoint, PrintingUnit.Inch)
		End If

	End Sub
	Private Function mFindExistingLegendItem(ByRef oFoundItem As CustomLegendItem, oLayer As Layer, oItemsToSearch As Collection(Of LegendItem)) As Boolean
		Dim bFoundIt As Boolean = False
		If oItemsToSearch IsNot Nothing Then
			For Each oCustomItem As CustomLegendItem In oItemsToSearch
				If oCustomItem.Layer.Equals(oLayer) Then
					oFoundItem = oCustomItem
					bFoundIt = True
					Exit For
				End If
			Next oCustomItem
		End If
		Return bFoundIt
	End Function

	Private Function mGetLegendWidth(ByVal oLegend As LegendPrinterLayer) As Single
		Dim dMaxWidth As Single = 0
		Dim dWidth As Single
		Dim oLegendItem As CustomLegendItem
		Dim oSizeF As SizeF
		Dim dTextHeight As Single
		Dim dImageHeight As Single
		Dim oLegendTitle As LegendItem
		Dim dLegendTitleTextWidth As Single

		'***** Assume we're dealing in pixels
		moGraphics.PageUnit = GraphicsUnit.Pixel

		oLegendTitle = CType(oLegend.Title, CustomLegendItem)
		With oLegendTitle
			oSizeF = moGraphics.MeasureString(.TextStyle.TextColumnName, DisplayUtilities.FontFromGeoFont(.TextStyle.Font))
			dMaxWidth = oSizeF.Width + .ImageWidth + .ImageLeftPadding + .ImageRightPadding + .TextLeftPadding + .TextRightPadding + .LeftPadding + .RightPadding
			dLegendTitleTextWidth = oSizeF.Width
			dTextHeight = oSizeF.Height + .TextTopPadding + .TextBottomPadding
			dImageHeight = .ImageHeight + .ImageTopPadding + .ImageBottomPadding
			If dTextHeight > dImageHeight Then
				.Height = dTextHeight
			Else
				.Height = dImageHeight
			End If
		End With

		For Each oLegendItem In oLegend.LegendItems
			With oLegendItem
				oSizeF = moGraphics.MeasureString(.TextStyle.TextColumnName, DisplayUtilities.FontFromGeoFont(.TextStyle.Font))
				dWidth = oSizeF.Width + .ImageWidth + .ImageLeftPadding + .ImageRightPadding + .TextLeftPadding + .TextRightPadding + .LeftPadding + .RightPadding
				'dWidth = oSizeF.Width + .ImageWidth + .ImageLeftPadding + .ImageRightPadding
			End With
			dTextHeight = oSizeF.Height + oLegendItem.TextTopPadding + oLegendItem.TextBottomPadding
			dImageHeight = oLegendItem.ImageHeight + oLegendItem.ImageTopPadding + oLegendItem.ImageBottomPadding
			If dTextHeight > dImageHeight Then
				oLegendItem.Height = dTextHeight
			Else
				oLegendItem.Height = dImageHeight
			End If

			If dWidth > dMaxWidth Then
				dMaxWidth = dWidth
			End If
		Next oLegendItem



		If dMaxWidth < (1.5 * msngDpi) Then
			dMaxWidth = CSng(1.5) * msngDpi
		End If

		With oLegendTitle
			.ImageHeight = 0
			.ImageWidth = 0
			.ImageBottomPadding = 0
			.ImageTopPadding = 0
			.ImageLeftPadding = 0
			.ImageRightPadding = 0
			.LeftPadding = ((dMaxWidth - dLegendTitleTextWidth) / 2)
		End With

		For Each oLegendItem In oLegend.LegendItems
			oLegendItem.Width = dMaxWidth
		Next oLegendItem

		Return dMaxWidth
	End Function
	Private Function mGetLegendHeight(ByVal oLegend As CustomLegendLayer) As Single
		Dim oLegendItem As CustomLegendItem
		Dim dHeight As Single

		oLegendItem = CType(oLegend.Title, CustomLegendItem)
		dHeight = oLegendItem.Height + oLegendItem.TopPadding + oLegendItem.BottomPadding

		For Each oLegendItem In oLegend.LegendItems
			dHeight += oLegendItem.Height + oLegendItem.TopPadding + oLegendItem.BottomPadding
		Next oLegendItem

		If oLegend.LegendItems.Count = 1 Then
			dHeight += 5
		End If

		Return dHeight + 3
	End Function
	Private Sub mSetPosition(ByVal dWidth As Double,
														ByVal dHeight As Double,
														ByVal oPrinterLayer As PrinterLayer,
														ByVal eInitialPosition As EnumPagePosition)
		Dim dCenterPointX As Double
		Dim dCenterPointY As Double
		Dim dDistanceFromEdge As Double = 0.25

		Dim oPageLayoutMap As MapPrinterLayer = ActiveLayoutMap()

		Dim oPrinterPage As CustomPagePrinterLayer = mGetPrinterPage()

		'***** Get the bounding box for the MapPrinterLayer
		'oPageLayoutMap.Open()
		oPrinterPage.Open()
		Dim oPageBoundingBox As RectangleShape = oPageLayoutMap.GetPosition(PrintingUnit.Inch)
		'''''''Dim oPageBoundingBox As RectangleShape = oPrinterPage.GetPosition(PrintingUnit.Inch)
		Select Case eInitialPosition
			Case EnumPagePosition.LowerRight
				dCenterPointX = oPageBoundingBox.LowerRightPoint.X - (dWidth / 2) - dDistanceFromEdge
				dCenterPointY = oPageBoundingBox.LowerRightPoint.Y + (dHeight / 2) + dDistanceFromEdge
			Case EnumPagePosition.LowerLeft
				dCenterPointX = oPageBoundingBox.LowerLeftPoint.X + (dWidth / 2) + dDistanceFromEdge
				dCenterPointY = oPageBoundingBox.LowerLeftPoint.Y + (dHeight / 2) + dDistanceFromEdge
			Case EnumPagePosition.TopRight
				dCenterPointX = oPageBoundingBox.UpperRightPoint.X - (dWidth / 2) - dDistanceFromEdge
				dCenterPointY = oPageBoundingBox.UpperRightPoint.Y - (dHeight / 2) - dDistanceFromEdge
			Case EnumPagePosition.TopLeft
				dCenterPointX = oPageBoundingBox.UpperLeftPoint.X + (dWidth / 2) + dDistanceFromEdge
				dCenterPointY = oPageBoundingBox.UpperLeftPoint.Y - (dHeight / 2) - dDistanceFromEdge
			Case EnumPagePosition.Current
				dCenterPointX = oPrinterLayer.GetPosition(PrintingUnit.Inch).GetCenterPoint.X
				dCenterPointY = oPrinterLayer.GetPosition(PrintingUnit.Inch).GetCenterPoint.Y
		End Select

		'***** Now let's try to place this layer in the lower right corner of the page
		oPrinterLayer.SetPosition(dWidth, dHeight, dCenterPointX, dCenterPointY, PrintingUnit.Inch)
	End Sub
	Public Sub RefreshMapExtents()
		Dim abIsDrawing() As Boolean = Nothing
		Dim iMapIndex As Integer
		Dim oPageLayoutMap As CustomMapPrinterLayer = Nothing
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = GetPageLayoutMaps()

		'***** This sub syncs the scale and extent of the layout maps
		'***** to that of the parent maps with which they are paired
		If Not oPageLayoutMaps Is Nothing Then
			ReDim abIsDrawing(oPageLayoutMaps.Count - 1)
			iMapIndex = 0
			For Each oPageLayoutMap In oPageLayoutMaps
				'***** Save the IsDrawing property on each PageLayoutMap
				abIsDrawing(iMapIndex) = oPageLayoutMap.IsDrawing
				iMapIndex += 1
				'***** Only sync the extent with the parent map if the
				'***** property is set to do so
				If oPageLayoutMap.SyncScaleWithParentMap Then
					oPageLayoutMap.Open()
					'***** Now set the extent on the emap
					oPageLayoutMap.MapExtent = mMapExtentFromScale(oPageLayoutMap)
					'***** Set the IsDrawing property, forcing all the layers to
					'***** redraw
					oPageLayoutMap.IsDrawing = True
				End If
			Next oPageLayoutMap
		End If

		'***** Call refresh on the printer interactive overlay
		Try
			Refresh(oPrinterOverlay)
		Catch ex As Exception
			MessageBox.Show("Error attempting to refresh layout map: " & ex.Message, "ERROR", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
		End Try

		'***** Set the IsDrawing property back to its original value
		If Not oPageLayoutMaps Is Nothing Then
			iMapIndex = 0
			For Each oPageLayoutMap In oPageLayoutMaps
				oPageLayoutMap.IsDrawing = abIsDrawing(iMapIndex)
				iMapIndex += 1
			Next oPageLayoutMap
		End If

	End Sub
	Private Function mMapExtentFromScale(oPageLayoutMap As CustomMapPrinterLayer) As RectangleShape
		Dim oMap As CustomWinformsMap = oPageLayoutMap.ParentMap
		Dim dScale As Double = oMap.CurrentScale
		Dim oWorldExtent As RectangleShape = oMap.CurrentExtent
		Dim oWorldExtentUnit As GeographyUnit = oMap.MapUnit
		Dim oPagePosition As RectangleShape = oPageLayoutMap.GetPosition
		Dim snglPageWidth As Single = CSng(oPagePosition.Width)
		Dim snglPageHeight As Single = CSng(oPagePosition.Height)
		Dim oExtent As RectangleShape

		oExtent = ExtentHelper.ZoomToScale(dScale, oWorldExtent, oWorldExtentUnit, snglPageWidth, snglPageHeight)
		Return oExtent

	End Function
	Public Sub RefreshPageLayoutMap(Optional ByVal oMap As CustomWinformsMap = Nothing)
		Dim bIsDrawing As Boolean
		Dim iMapIndex As Integer
		Dim oPageLayoutMap As CustomMapPrinterLayer = Nothing
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer)

		If oMap Is Nothing Then
			oPageLayoutMaps = GetPageLayoutMaps()
		Else
			oPageLayoutMaps = New Collection(Of CustomMapPrinterLayer)
			oPageLayoutMaps.Add(mLayoutMapFromParentMap(oMap))
		End If

		If Not oPageLayoutMaps Is Nothing Then
			iMapIndex = 0
			For Each oPageLayoutMap In oPageLayoutMaps
				oPageLayoutMap.Open()
				bIsDrawing = oPageLayoutMap.IsDrawing
				iMapIndex += 1
				'***** Set the IsDrawing property, forcing all the layers to
				'***** redraw
				oPageLayoutMap.IsDrawing = True
			Next oPageLayoutMap
		End If

		'***** Call refresh on the printer interactive overlay
		Try
			'***** Re-query any data grids in case featues have moved in or out of the extent
			mRefreshDataGrids()
			Refresh(oPrinterOverlay)
		Catch ex As Exception
			MessageBox.Show("Error attempting to refresh layout map: " & ex.Message, "ERROR", MessageBoxButtons.OK, MessageBoxIcon.Exclamation)
			Debug.WriteLine("")
		End Try

		If Not oPageLayoutMaps Is Nothing Then
			For Each oPageLayoutMap In oPageLayoutMaps
				If Not oPageLayoutMap.Legend Is Nothing Then
					If oPageLayoutMap.Legend.NeedsRefresh Then
						SyncLegendToLayers(oPageLayoutMap.ParentMap)
					End If
				End If
			Next oPageLayoutMap
		End If

		'***** Set the IsDrawing property back to its original value
		If Not oPageLayoutMaps Is Nothing Then
			iMapIndex = 0
			For Each oPageLayoutMap In oPageLayoutMaps
				oPageLayoutMap.IsDrawing = bIsDrawing
				iMapIndex += 1
			Next oPageLayoutMap
		End If
		Debug.WriteLine(GetScale.ToString)
	End Sub
	Private Sub mRefreshDataGrids()
		Dim oPrinterOverLay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oDataGrid As CustomDataGridPrinterLayer

		For Each oPrinterLayer As PrinterLayer In oPrinterOverLay.PrinterLayers
			If TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
				oDataGrid = CType(oPrinterLayer, CustomDataGridPrinterLayer)
				mRefreshDataGrid(oDataGrid)
			End If
		Next oPrinterLayer
	End Sub
	Private Sub mRefreshDataGrid(oDataGrid As CustomDataGridPrinterLayer)
		Dim iNumRowsBeforeRefresh As Integer = 0
		Dim iNumRowsAfterRefresh As Integer = 0
		Dim dHeight As Double
		With oDataGrid
			If .DataTable IsNot Nothing Then
				iNumRowsBeforeRefresh = .DataTable.Rows.Count
			End If
			If Not .FillDataTable() Then
				Return
			End If
			If .DataTable IsNot Nothing Then
				iNumRowsAfterRefresh = .DataTable.Rows.Count
			End If
			Using oGraphics As System.Drawing.Graphics = Me.CreateGraphics
				dHeight = .GetGridHeight(oGraphics)
			End Using
			.Open()
		End With
		'***** If the number of rows in the datagrid has changed, reset it's size
		'***** but keep it in the same position
		If iNumRowsBeforeRefresh <> iNumRowsAfterRefresh Then
			mSetPosition(6, dHeight, oDataGrid, EnumPagePosition.Current)
			'***** Simulate MouseUp/MouseDown, resizes the grid to have the correct width
			'***** Something internal to the ThinkGeo library does this in response
			mSimulateMouseUpDown()
		End If

	End Sub
	Private Function mAskDeleteOK(oPrinterLayer As PrinterLayer) As Boolean
		Dim sQuestion As String = String.Empty
		Dim sCaption As String = String.Empty
		Dim bResult As Boolean = False

		'***** Ask the user if they want to delete an element from the layout map
		'***** Build the question based on what type of element they are deleting
		If TypeOf oPrinterLayer Is LegendPrinterLayer Then
			sQuestion = DeleteLegendQuestion
			sCaption = RemoveLegendCaption
		ElseIf TypeOf oPrinterLayer Is ScaleLinePrinterLayer Then
			sQuestion = DeleteScaleLineQuestion
			sCaption = RemoveScalelineCaption
		ElseIf TypeOf oPrinterLayer Is ImagePrinterLayer Then
			sQuestion = DeleteImageQuestion
			sCaption = RemoveImageCaption
		ElseIf TypeOf oPrinterLayer Is LabelPrinterLayer Then
			sQuestion = DeleteLabelQuestion
			sCaption = RemoveLabelCaption
		ElseIf TypeOf oPrinterLayer Is ScaleBarPrinterLayer Then
			sQuestion = DeleteScalebarQuestion
			sCaption = RemoveScalebarCaption
		ElseIf TypeOf oPrinterLayer Is DataGridPrinterLayer Then
			sQuestion = DeleteDataGridQuestion
			sCaption = RemoveDataGridCaption
		ElseIf TypeOf oPrinterLayer Is MagneticDeclinationPrinterLayer Then
			sQuestion = DeleteMagneticNorthArrowQuestion
			sCaption = RemoveMagneticNorthArrowCaption
		ElseIf TypeOf oPrinterLayer Is ScaleTextPrinterLayer Then
			sQuestion = DeleteScaleTextQuestion
			sCaption = RemoveScaletextCaption
		End If

		If Not sQuestion = String.Empty Then
			If MessageBox.Show(sQuestion, sCaption, MessageBoxButtons.YesNo, MessageBoxIcon.Question) = Windows.Forms.DialogResult.Yes Then
				'*****Return true, meaning go ahead with the delete
				bResult = True
			End If
		End If

		Return bResult

	End Function
	Public Sub SaveTemplate()
		Dim oSaveDlg As New SaveFileDialog
		Dim oGeoSerializer As GeoSerializer
		Dim sTempFolderName = System.IO.Path.GetTempPath
		Dim sTempZipFileName As String
		Dim oFilesToDelete As Collection(Of String)
		Dim sTemplateFileName As String = String.Empty
		With oSaveDlg
			.Filter = "Page Layout Templates (*.tpl)|*.tpl"
			If .ShowDialog() = Windows.Forms.DialogResult.Cancel Then
				Return
			End If
			oGeoSerializer = New GeoSerializer
			oFilesToDelete = New Collection(Of String)
			sTempZipFileName = sTempFolderName & Path.GetFileName(.FileName)
			If File.Exists(sTempZipFileName) Then
				File.Delete(sTempZipFileName)
			End If
			SerializeLayoutTemplate(oGeoSerializer, sTempZipFileName, sTempFolderName, oFilesToDelete)
			oGeoSerializer = Nothing
			sTemplateFileName = .FileName
		End With

		For Each sFile In oFilesToDelete
			File.Delete(sFile)
		Next sFile

		'***** Now move the file that was just created to its permanent location
		'***** Zip Up the files and put them where they go
		If File.Exists(sTemplateFileName) Then
			File.Delete(sTemplateFileName)
		End If

		File.Move(sTempZipFileName, sTemplateFileName)
	End Sub
	Public Sub SerializeLayoutTemplate(ByVal oGeoSerializer As GeoSerializer, sTempZipFileName As String, sFolder As String, oFilesToDelete As Collection(Of String))
		'***** Get the PrinterInteractiveOverlay and serialize it
		Dim oPrinterInteractiveOverlay As PrinterInteractiveOverLay = PrinterOverlay
		Dim sTempFileName As String = String.Empty
		Dim oPageLayoutMap As CustomMapPrinterLayer '= PageLayout.CurrentLayoutMap
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = GetPageLayoutMaps()
		Dim oCollectionOfLayerCollections As New Collection(Of Collection(Of Layer))
		Dim oLayers As Collection(Of Layer)
		Dim oPrinterLayer As PrinterLayer
		Dim sFileName As String

		If Not oPageLayoutMaps Is Nothing Then
			For Each oPageLayoutMap In oPageLayoutMaps
				oLayers = New Collection(Of Layer)
				oCollectionOfLayerCollections.Add(oLayers)
				'***** Save the layers from the map, we don't want to save them
				'***** as part of the template
				For i As Integer = 0 To oPageLayoutMap.Layers.Count - 1
					oLayers.Add(oPageLayoutMap.Layers(i))
				Next i
				'***** Now clear the map layers
				oPageLayoutMap.Layers.Clear()
			Next oPageLayoutMap
		End If

		For i = 0 To oPrinterInteractiveOverlay.PrinterLayers.Count - 1
			oPrinterLayer = oPrinterInteractiveOverlay.PrinterLayers(i)
			If TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
				Dim oCustDataGrid As CustomDataGridPrinterLayer = CType(oPrinterLayer, CustomDataGridPrinterLayer)
				If oCustDataGrid.ConnectionInfo IsNot Nothing Then
					oCustDataGrid.ConnectionInfo = oCustDataGrid.ConnectionInfo.Copy
					oCustDataGrid.ConnectionInfo.UserName = String.Empty
					oCustDataGrid.ConnectionInfo.Password = String.Empty
				End If
			End If
			oPrinterLayer.Close()
			sFileName = sFolder & "PrinterLayer_" & (i + 1).ToString & ".xml"
			oGeoSerializer.Serialize(oPrinterLayer, sFileName)
			'***** Add to the zip file
			ZipArchive.AddZipFile(sTempZipFileName, sFileName, FileMode.OpenOrCreate)
			'***** Delete the xml file
			'File.Delete(sFileName)
			oFilesToDelete.Add(sFileName)
		Next i

		Dim iLayersCollectionIdx As Integer = 0
		If Not oPageLayoutMaps Is Nothing Then
			For Each oPageLayoutMap In oPageLayoutMaps
				oLayers = oCollectionOfLayerCollections(iLayersCollectionIdx)
				'***** Save the layers from the map, we don't want to save them
				'***** as part of the template
				For i As Integer = 0 To oLayers.Count - 1
					oPageLayoutMap.Layers.Add(oLayers(i))
				Next i
				iLayersCollectionIdx += 1
			Next oPageLayoutMap
		End If

	End Sub
	Public Sub SerializeLayoutTemplate_Old(Optional oTemplateStream As Stream = Nothing)
		'***** Get the PrinterInteractiveOverlay and serialize it
		Dim oFormatter As BinaryFormatter
		Dim oStream As Stream = Nothing
		Dim oPrinterInteractiveOverlay As PrinterInteractiveOverLay = PrinterOverlay
		Dim sTempFileName As String = String.Empty
		Dim oPageLayoutMap As CustomMapPrinterLayer '= PageLayout.CurrentLayoutMap
		Dim oPageLayoutMaps As Collection(Of CustomMapPrinterLayer) = GetPageLayoutMaps()
		Dim oCollectionOfLayerCollections As New Collection(Of Collection(Of Layer))
		Dim oLayers As Collection(Of Layer)
		Dim oPrinterLayer As PrinterLayer

		If Not oPageLayoutMaps Is Nothing Then
			For Each oPageLayoutMap In oPageLayoutMaps
				oLayers = New Collection(Of Layer)
				oCollectionOfLayerCollections.Add(oLayers)
				'***** Save the layers from the map, we don't want to save them
				'***** as part of the template
				For i As Integer = 0 To oPageLayoutMap.Layers.Count - 1
					oLayers.Add(oPageLayoutMap.Layers(i))
				Next i
				'***** Now clear the map layers
				oPageLayoutMap.Layers.Clear()
			Next oPageLayoutMap
		End If

		'***** Open a file stream to use for serialization
		oFormatter = New BinaryFormatter()

		'***** If a stream is passed in we don't want to create a new one
		If oTemplateStream Is Nothing Then

			If Not DotNetUtils.GetSaveFileName("Template Files (*.tpl)|*.tpl", TemplateFileName) Then
				Return
			End If

			sTempFileName = System.IO.Path.GetTempFileName()
			'***** We will write to a temp file so in case something goes wrong
			'***** the original file (if there was one) will still be OK.
			oStream = New FileStream(sTempFileName, FileMode.Create, FileAccess.Write, FileShare.None)
		Else
			oStream = oTemplateStream
		End If

		For i = 0 To oPrinterInteractiveOverlay.PrinterLayers.Count - 1
			oPrinterLayer = oPrinterInteractiveOverlay.PrinterLayers(i)
			oFormatter.Serialize(oStream, oPrinterLayer)
		Next i

		Dim iLayersCollectionIdx As Integer = 0
		If Not oPageLayoutMaps Is Nothing Then
			For Each oPageLayoutMap In oPageLayoutMaps
				oLayers = oCollectionOfLayerCollections(iLayersCollectionIdx)
				'***** Save the layers from the map, we don't want to save them
				'***** as part of the template
				For i As Integer = 0 To oLayers.Count - 1
					oPageLayoutMap.Layers.Add(oLayers(i))
				Next i
				iLayersCollectionIdx += 1
			Next oPageLayoutMap
		End If

		'***** Close the stream so we can move the file
		If oTemplateStream Is Nothing Then
			oStream.Close()
		Else
			'***** A stream was passed in by the calling function so
			'***** we're done.
			Return
		End If

		'***** Delete the original file if it exists
		If File.Exists(TemplateFileName) Then
			File.Delete(TemplateFileName)
		End If

		'***** Now move the file that was just created to its permanent location
		File.Move(sTempFileName, TemplateFileName)

		'***** If you don't add the group 'Authenticated Users' with some rights
		'***** the file has a little padlock icon on it; that can't be a good thing
		'***** so here we grant access to 'Authenticated Users'
		DotNetUtils.AddFileSecurity(TemplateFileName, "Authenticated Users", FileSystemRights.Modify, AccessControlType.Allow)
		DotNetUtils.AddFileSecurity(TemplateFileName, "Authenticated Users", FileSystemRights.ReadAndExecute, AccessControlType.Allow)
		DotNetUtils.AddFileSecurity(TemplateFileName, "Authenticated Users", FileSystemRights.Read, AccessControlType.Allow)
		DotNetUtils.AddFileSecurity(TemplateFileName, "Authenticated Users", FileSystemRights.Write, AccessControlType.Allow)

	End Sub
	Public ReadOnly Property LayoutMode As PageLayoutControl.EnumPageLayoutMode
		Get
			If PageLayoutControl IsNot Nothing Then
				Return PageLayoutControl.PageLayoutMode
			Else
				Return MapUtilities.PageLayoutControl.EnumPageLayoutMode.TemplatePreview
			End If
		End Get
	End Property
	Public Sub LoadPageLayoutTemplate(oMapControl As SilvicsMapControl,
																		 Optional sTemplateFileName As String = "",
																		 Optional oExtent As RectangleShape = Nothing,
																		 Optional dScale As Double = -1,
																		 Optional oAutoPageLayoutPrinterLayers As Collection(Of PrinterLayer) = Nothing,
																		 Optional oLayersInLegend As Collection(Of Layer) = Nothing,
																		 Optional oFeature As Feature = Nothing)

		'***** When this sub is called with all the optional parameters
		'***** it is being used to load a page layout that is responding
		'***** to an auto-layout request. 
		'***** SET 01/21/2016

		Dim oPrinterLayers As Collection(Of PrinterLayer) = Nothing
		Dim sKey As String = String.Empty
		Dim oBrush As New GeoSolidBrush(GeoColor.SimpleColors.Transparent)
		Dim bSuccess As Boolean = False

		If oAutoPageLayoutPrinterLayers Is Nothing Then
			'***** We are loading up a regular page layout map
			oPrinterLayers = PageLayoutHelper.LoadPrinterLayersFromTemplateFile(sTemplateFileName)
		Else
			'***** An Auto layout is being loaded
			oPrinterLayers = oAutoPageLayoutPrinterLayers
		End If

		If oPrinterLayers Is Nothing Then
			Return
		End If

		If LayoutMode <> MapUtilities.PageLayoutControl.EnumPageLayoutMode.AutoLayout Then
			ClearPageLayout()
		End If

		CreateDefaultLayout()

		If LayoutMode = MapUtilities.PageLayoutControl.EnumPageLayoutMode.AutoLayout Then
			CType(PrinterOverlay, CustomPrinterOverlay).AutoLayoutFeature = oFeature
		End If

		If Not ActiveLayoutMap Is Nothing Then
			ActiveLayoutMap.BackgroundMask = DisplayUtilities.CreateSimpleAreaStyle(oBrush.Color)
		End If

		If Not oPrinterLayers Is Nothing Then
			'***** Clear any printer layers then add the layers from the collection
			PrinterOverlay.PrinterLayers.Clear()
			For i = 0 To oPrinterLayers.Count - 1 'To 0 Step -1
				'QB #2339, Jamie Irwin 2/18/2016, Added some exception code to handle opening an older Template.
				'This is Modeled after code written by Steve Teller in WorkingSet.OpenWorkingSet at or about line #382
				If TypeOf oPrinterLayers(i) Is PagePrinterLayer Then
					sKey = SilvicsPageLayout.csPageLayerKey
					If TypeOf oPrinterLayers(i) Is CustomPagePrinterLayer Then
						PrinterOverlay.PrinterLayers.Add(sKey, oPrinterLayers(i))
					Else
						'***** This is an old PagePrinterLayer, before the CustomPagePrinterLayer came along,
						'***** So create a New CustomPagePrinterLayer with its properties
						Dim oCustomPagePrinterLayer As New CustomPagePrinterLayer(CType(oPrinterLayers(i), PagePrinterLayer))
						PrinterOverlay.PrinterLayers.Add(sKey, oCustomPagePrinterLayer)
					End If
				ElseIf TypeOf oPrinterLayers(i) Is CustomPagePrinterLayer Then
					sKey = SilvicsPageLayout.csPageLayerKey
					PrinterOverlay.PrinterLayers.Add(sKey, CType(oPrinterLayers(i).CloneDeep, CustomPagePrinterLayer))
				ElseIf TypeOf oPrinterLayers(i) Is CustomMapPrinterLayer Then
					Dim oMapPrinterLayer = CType(oPrinterLayers(i), CustomMapPrinterLayer).Clone
					oMapPrinterLayer.IsDrawing = False
					sKey = SilvicsPageLayout.csMapPrinterLayerBaseKey & "_" & oMapPrinterLayer.PageLayoutMapNumber
					PrinterOverlay.PrinterLayers.Add(sKey, oMapPrinterLayer)
				ElseIf TypeOf oPrinterLayers(i) Is ScaleLinePrinterLayer Then
					Dim oScaleLineLayer = CType(oPrinterLayers(i).CloneDeep, ScaleLinePrinterLayer)
					sKey = oScaleLineLayer.Name
					PrinterOverlay.PrinterLayers.Add(sKey, oScaleLineLayer)
				ElseIf TypeOf oPrinterLayers(i) Is ScaleBarPrinterLayer Then
					Dim oScalebarLayer As ScaleBarPrinterLayer = CType(oPrinterLayers(i).CloneDeep, ScaleBarPrinterLayer)
					sKey = oScalebarLayer.Name
					If oScalebarLayer.BarBrush Is Nothing Then
						oScalebarLayer.BarBrush = New GeoSolidBrush(GeoColor.SimpleColors.Black)
					End If
					PrinterOverlay.PrinterLayers.Add(sKey, oScalebarLayer)
				ElseIf TypeOf oPrinterLayers(i) Is CustomLegendLayer Then
					'Dim oLegendLayer As CustomLegendLayer = CType(oPrinterLayers(i), CustomLegendLayer).Clone
					Dim oLegendLayer As New CustomLegendLayer(CType(oPrinterLayers(i), CustomLegendLayer))
					''''''''oLegendLayer.SilvicsPageLayoutMap = Me
					sKey = oLegendLayer.Name
					oLegendLayer.LegendItems.Clear()
					PrinterOverlay.PrinterLayers.Add(sKey, oLegendLayer)
				ElseIf TypeOf oPrinterLayers(i) Is CustomMagDecNorthArrow Then
					Dim oMagDecNorthArrow As CustomMagDecNorthArrow = CType(oPrinterLayers(i), CustomMagDecNorthArrow).Clone
					sKey = oMagDecNorthArrow.Name
					PrinterOverlay.PrinterLayers.Add(sKey, oMagDecNorthArrow)
				ElseIf TypeOf oPrinterLayers(i) Is CustomDataGridPrinterLayer Then
					bSuccess = True
					Dim oCustomDataGrid As CustomDataGridPrinterLayer = CType(oPrinterLayers(i), CustomDataGridPrinterLayer).Clone
					sKey = oCustomDataGrid.Name
					If StandardLayer IsNot Nothing Then
						Try
							oCustomDataGrid.AppendSingleFeatureWhereClause(StandardLayer, mGetAutoLayoutFeature)
						Catch ex As SqlClient.SqlException
							bSuccess = False
							Dim sMsg As String = My.Resources.CustomDataGridPrinterLayerResources.MsgProblemWithDbConnection
							Dim sWhatHappened As String = My.Resources.CustomDataGridPrinterLayerResources.MsgDataGridNotAdded
							ErrHandler.ShowDialog(sMsg, sWhatHappened, ErrHandler.OKToContinue, ex)
						Catch ex As Exception
							bSuccess = False
							Dim sMsg As String = My.Resources.CustomDataGridPrinterLayerResources.MsgPossibleQueryError
							Dim sWhatHappened As String = My.Resources.CustomDataGridPrinterLayerResources.MsgDataGridNotAdded
							ErrHandler.ShowDialog(sMsg, sWhatHappened, ErrHandler.OKToContinue, ex)
						End Try
					End If
					If bSuccess Then
						oCustomDataGrid.InTemplateMode = False
						If oCustomDataGrid.FillDataTable() Then
							PrinterOverlay.PrinterLayers.Add(sKey, oCustomDataGrid)
						End If
					End If
				Else
					Dim oPrinterLayer As PrinterLayer = CType(oPrinterLayers(i).CloneDeep, PrinterLayer)
					PrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oPrinterLayer)
				End If
			Next i
		End If

		'***** Here is where we load the map layers into the page layout map
		'***** If we are only looking at a template preview we don't want to 
		'***** load the layers
		Dim oPageLayoutMaps As New Collection(Of CustomMapPrinterLayer)
		If LayoutMode <> MapUtilities.PageLayoutControl.EnumPageLayoutMode.TemplatePreview Then
			For Each oMapPrinterLayer As CustomMapPrinterLayer In GetPageLayoutMaps()
				For Each oMap In oMapControl.GetMaps
					AddPageLayoutMap(oMap)
					If oMapPrinterLayer.PageLayoutMapNumber = oMap.PageLayoutMapNumber Then
						With oMapPrinterLayer
							.MapUnit = oMap.MapUnit
							.ParentMap = oMap
							If oExtent IsNot Nothing Then
								If dScale > 0 Then
									oExtent = PageLayoutHelper.MapExtentFromScale(dScale, oExtent, .MapUnit, oMapPrinterLayer, EnumScaleMode.SpecificScale)
									.MapExtent = oExtent
								Else
									.MapExtent = oExtent
									Dim dCurrentScale As Double = .CurrentScale
									'***** Get the next larger scale
									dScale = ZoomLevelHelper.GetNextHigherScale(dCurrentScale)
									.MapExtent = PageLayoutHelper.MapExtentFromScale(dScale, oExtent, .MapUnit, oMapPrinterLayer, EnumScaleMode.BestFit)
								End If
								.SyncScaleWithParentMap = False
							End If
							oPageLayoutMaps.Add(oMapPrinterLayer)
						End With
						Dim oNonGroupLayers = oMap.GetAllNonGroupLayers
						If oNonGroupLayers IsNot Nothing Then
							For Each oLayer As Layer In oNonGroupLayers
								oMapPrinterLayer.Layers.Add(oLayer)
							Next oLayer
							Exit For
						End If
					End If
				Next oMap
			Next oMapPrinterLayer
		End If

		'***** Pair each map element with its parent PageLayoutMap (when necessary)
		SetMapPrinterLayerForElements()

		'***** If the map has a legend, add the layers selected by the user
		For Each oMapPrinterLayer In oPageLayoutMaps
			If oMapPrinterLayer.Legend IsNot Nothing Then
				If oLayersInLegend IsNot Nothing Then
					RefreshLegendLayers(oLayersInLegend, oMapPrinterLayer.Legend, Enums.EnumRefreshBehavior.DoNotRefresh)
				Else
					Dim oGeoLayers = oMapPrinterLayer.ParentMap.GetAllNonGroupLayers
					Dim oLayers As New Collection(Of Layer)

					For i = oGeoLayers.Count - 1 To 0 Step -1
						oLayers.Add(oGeoLayers(i))
					Next i

					RefreshLegendLayers(oLayers, oMapPrinterLayer.Legend, Enums.EnumRefreshBehavior.DoNotRefresh)
				End If
			End If
		Next oMapPrinterLayer

		'***** Set the combo boxes on the PageLayoutControl to reflect the options
		'***** on this map
		PageLayoutControl.SetCombosFromPageLayout()
		If LayoutMode <> MapUtilities.PageLayoutControl.EnumPageLayoutMode.AutoLayout Then
			RefreshMapExtents()
		End If
	End Sub
	Public Function LoadPrinterLayersFromTemplateFolder(sFolder As String) As Collection(Of PrinterLayer)
		Dim oPrinterLayer As Object
		Dim oPrinterLayers As New Collection(Of PrinterLayer)
		Dim oGeoSerializer As New GeoSerializer
		Dim iPrinterLayerNum As Integer
		Dim sFileName As String

		'***** The xml files for the printer layers
		'***** have already been unzipped
		'***** Get file to open
		iPrinterLayerNum = 1
		sFileName = sFolder & "PrinterLayer_" & iPrinterLayerNum.ToString & ".xml"
		While File.Exists(sFileName)
			oPrinterLayer = oGeoSerializer.Deserialize(sFileName, GeoFileReadWriteMode.Read)
			If TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
				'***** QB #2452
				'***** SET 09/16/2016
				'***** The username and password do not get saved when the DataGrid is saved,
				'***** so they must be inserted here.
				Dim oCustDbGrid As CustomDataGridPrinterLayer = CType(oPrinterLayer, CustomDataGridPrinterLayer)
				If oCustDbGrid.ConnectionInfo Is Nothing Then
					oCustDbGrid.ConnectionInfo = Utils.CurrentPlugin.PrimaryDatabaseConnectionInfo
				End If
				oCustDbGrid.ConnectionInfo = DatabaseUtils.SqlConnectionInfoFromExisting(oCustDbGrid.ConnectionInfo)
			End If
			oPrinterLayers.Add(CType(oPrinterLayer, PrinterLayer))
			iPrinterLayerNum += 1
			IO.File.Delete(sFileName)
			sFileName = sFolder & "PrinterLayer_" & iPrinterLayerNum.ToString & ".xml"
		End While

		Return oPrinterLayers

	End Function
	'Public Function LoadPrinterLayersFromTemplateFile(Optional ByRef rsFileName As String = "") As Collection(Of PrinterLayer)
	'   Dim oPrinterLayer As Object
	'   Dim oPrinterLayers As New Collection(Of PrinterLayer)
	'   Dim oGeoSerializer As New GeoSerializer
	'   Dim iPrinterLayerNum As Integer
	'   Dim sTempXmlFolder As String = IO.Path.GetTempPath

	'   If rsFileName = String.Empty Then
	'      If Not DotNetUtils.GetDocumentFileName("Open Layout Template", "Template Files (*.tpl)|*.tpl", rsFileName) Then
	'         Return Nothing
	'      End If
	'   End If

	'   TemplateFileName = rsFileName
	'   Try
	'      Using oZipArchive = ZipArchive.OpenOnFile(TemplateFileName)
	'         Dim oFiles = oZipArchive.Files
	'         For Each oFile As ZipArchive.ZipFileInfo In oFiles
	'            ZipArchive.UnZipFile(TemplateFileName, sTempXmlFolder & oFile.Name)
	'         Next oFile
	'      End Using
	'   Catch ex As System.Reflection.TargetInvocationException

	'   End Try

	'   iPrinterLayerNum = 1
	'   Dim sFileName = sTempXmlFolder & "PrinterLayer_" & iPrinterLayerNum.ToString & ".xml"
	'   While File.Exists(sFileName)
	'      oPrinterLayer = oGeoSerializer.Deserialize(sFileName, FileAccess.Read)
	'      oPrinterLayers.Add(CType(oPrinterLayer, PrinterLayer))
	'      iPrinterLayerNum += 1
	'      IO.File.Delete(sFileName)
	'      sFileName = sTempXmlFolder & "PrinterLayer_" & iPrinterLayerNum.ToString & ".xml"
	'   End While

	'   Return oPrinterLayers

	'End Function

	Public Function LoadPrinterLayersFromTemplateFile_Old(Optional oTemplateStream As FileStream = Nothing) As Collection(Of PrinterLayer)
		Dim oFormatter As New BinaryFormatter
		Dim oStream As FileStream
		Dim oPrinterLayer As Object
		Dim oPrinterLayers As New Collection(Of PrinterLayer)

		If oTemplateStream Is Nothing Then
			'***** Get file to open
			If Not DotNetUtils.GetDocumentFileName("Open Layout Template", "Template Files (*.tpl)|*.tpl", TemplateFileName) Then
				Return Nothing
			End If

			'***** Open a file stream to use for deserialization
			oStream = New FileStream(TemplateFileName, FileMode.Open, FileAccess.Read, FileShare.Read)
		Else
			'***** Use the stream provided by the caller
			oStream = oTemplateStream
		End If

		While oStream.Position < oStream.Length
			oPrinterLayer = oFormatter.Deserialize(oStream)
			oPrinterLayers.Add(CType(oPrinterLayer, PrinterLayer))
		End While

		'***** If the caller did not provide a stream, close it
		If oTemplateStream Is Nothing Then
			oStream.Close()
		End If


		'LoadLayoutFromTemplate(oPrinterLayers, CurrentParentMap)
		Return oPrinterLayers

	End Function
	'Public Sub LoadLayoutFromTemplate(oPrinterLayers As Collection(Of PrinterLayer), oMap As CustomWinformsMap)
	'   Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()

	'   '***** Clear the printer layers
	'   oPrinterOverlay.PrinterLayers.Clear()
	'   'RefreshMapLayers()

	'   For Each oPrinterLayer In oPrinterLayers
	'      If TypeOf oPrinterLayer Is CustomPagePrinterLayer Then
	'         '***** Create the CustomPagePrinterLayer which shows the page boundary and 
	'         '***** is the area the user will arrange all of the other layers on top of.
	'         oPrinterOverlay.PrinterLayers.Add(csPageLayerKey, oPrinterLayer)
	'      ElseIf TypeOf oPrinterLayer Is CustomMapPrinterLayer Then
	'         oPrinterOverlay.PrinterLayers.Add(oPrinterLayer.Name, oPrinterLayer)
	'         '***** For some reason 'IsDrawing' must be set to True or a weird
	'         '***** exception is generated when you try to refresh the Page Layout
	'         oPrinterLayer.IsDrawing = True
	'      ElseIf TypeOf oPrinterLayer Is ScaleLinePrinterLayer Then
	'         oPrinterOverlay.PrinterLayers.Add(oPrinterLayer.Name, oPrinterLayer)
	'      ElseIf TypeOf oPrinterLayer Is ScaleBarPrinterLayer Then
	'         oPrinterOverlay.PrinterLayers.Add(oPrinterLayer.Name, oPrinterLayer)
	'      ElseIf TypeOf oPrinterLayer Is CustomLegendLayer Then
	'         oPrinterOverlay.PrinterLayers.Add(oPrinterLayer.Name, oPrinterLayer)
	'      ElseIf TypeOf oPrinterLayer Is ImagePrinterLayer Then
	'         'Debug.WriteLine("Image Layer")
	'         oPrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oPrinterLayer)
	'      Else
	'         oPrinterOverlay.PrinterLayers.Add(Guid.NewGuid.ToString, oPrinterLayer)
	'      End If
	'   Next oPrinterLayer

	'   '***** Load the map layers into the template
	'   LoadTemplateFromMap(oMap)

	'   ViewEntirePageLayout()

	'End Sub
	Public Sub SetMapPrinterLayerForElements()
		'***** Scalebar layers and scaleline layer must be paired with their page layout maps
		'***** when they are opened from a Working Set
		Dim oPrinterLayers As GeoCollection(Of PrinterLayer) = mGetPrinterOverlay.PrinterLayers
		For Each oPrinterLayer As PrinterLayer In oPrinterLayers
			If TypeOf oPrinterLayer Is ScaleLinePrinterLayer Then
				mPairScaleLineWithPageLayout(CType(oPrinterLayer, ScaleLinePrinterLayer), oPrinterLayers)
			ElseIf TypeOf oPrinterLayer Is ScaleBarPrinterLayer Then
				mPairScaleBarWithPageLayout(CType(oPrinterLayer, ScaleBarPrinterLayer), oPrinterLayers)
			ElseIf TypeOf oPrinterLayer Is CustomLegendLayer Then
				mSetLegendOnPageLayout(CType(oPrinterLayer, CustomLegendLayer), oPrinterLayers)
			ElseIf TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
				mPairDataGridWithPageLayout(CType(oPrinterLayer, CustomDataGridPrinterLayer), oPrinterLayers)
			ElseIf TypeOf oPrinterLayer Is CustomMagDecNorthArrow Then
				mPairMagDecNorthArrowWithPageLayout(CType(oPrinterLayer, CustomMagDecNorthArrow), oPrinterLayers)
			ElseIf TypeOf oPrinterLayer Is ScaleTextPrinterLayer Then
				mPairScaleTextWithPageLayout(CType(oPrinterLayer, ScaleTextPrinterLayer), oPrinterLayers)
			End If
		Next oPrinterLayer
	End Sub
	Private Sub mSetLegendOnPageLayout(oLegend As CustomLegendLayer, oPrinterLayers As GeoCollection(Of PrinterLayer))
		Dim sPageLayoutKey As String = mGetPageLayoutKeyFromElementKey(oLegend.Name)
		CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer).Legend = oLegend
		'***** Set two key properties on the Legend
		'***** SET QB #3642
		'***** 08/08/2017
		oLegend.SilvicsPageLayout = Me
		oLegend.PageLayoutMap = CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer)
	End Sub
	Private Sub mPairScaleBarWithPageLayout(oScaleBar As ScaleBarPrinterLayer, oPrinterLayers As GeoCollection(Of PrinterLayer))
		Dim sPageLayoutKey As String = mGetPageLayoutKeyFromElementKey(oScaleBar.Name)
		Dim oMapPrinterLayer As CustomMapPrinterLayer = CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer)
		oScaleBar.MapPrinterLayer = oMapPrinterLayer
		oScaleBar.MapUnit = oMapPrinterLayer.MapUnit
	End Sub
	Private Sub mPairMagDecNorthArrowWithPageLayout(oMagDecNorthArrow As CustomMagDecNorthArrow, oPrinterLayers As GeoCollection(Of PrinterLayer))
		Dim sPageLayoutKey As String = mGetPageLayoutKeyFromElementKey(oMagDecNorthArrow.Name)
		Dim oMapPrinterLayer As CustomMapPrinterLayer = CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer)
		oMagDecNorthArrow.MapPrinterLayer = oMapPrinterLayer
		oMagDecNorthArrow.SetCoordinateSystem()
	End Sub
	Private Sub mPairScaleTextWithPageLayout(oScaleText As ScaleTextPrinterLayer, oPrinterLayers As GeoCollection(Of PrinterLayer))
		Dim sPageLayoutKey As String = mGetPageLayoutKeyFromElementKey(oScaleText.Name)
		Dim oMapPrinterLayer As CustomMapPrinterLayer = CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer)
		oScaleText.MapPrinterLayer = oMapPrinterLayer
	End Sub

	Private Sub mPairScaleLineWithPageLayout(oScaleLine As ScaleLinePrinterLayer, oPrinterLayers As GeoCollection(Of PrinterLayer))
		Dim sPageLayoutKey As String = mGetPageLayoutKeyFromElementKey(oScaleLine.Name)
		Dim oMapPrinterLayer As CustomMapPrinterLayer = CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer)
		oScaleLine.MapPrinterLayer = oMapPrinterLayer
		oScaleLine.MapUnit = oMapPrinterLayer.MapUnit
	End Sub
	Private Sub mPairDataGridWithPageLayout(oDataGrid As CustomDataGridPrinterLayer, oPrinterLayers As GeoCollection(Of PrinterLayer))
		Dim sPageLayoutKey As String = mGetPageLayoutKeyFromElementKey(oDataGrid.Name)
		oDataGrid.MapPrinterLayer = CType(oPrinterLayers(sPageLayoutKey), CustomMapPrinterLayer)
	End Sub

	Private Function mGetPageLayoutKeyFromElementKey(sElementKey As String) As String
		'***** Get the PageLayoutNumber from the element key
		Dim sPageLayoutNumber As String = sElementKey.Substring(sElementKey.IndexOf("_"))
		If sPageLayoutNumber.IndexOf("_", 1) <> -1 Then
			sPageLayoutNumber = sPageLayoutNumber.Substring(0, sPageLayoutNumber.IndexOf("_", 1))
		End If
		Return csMapPrinterLayerBaseKey & sPageLayoutNumber
	End Function

	Public Sub LoadTemplateFromMap(oMap As CustomWinformsMap)
		'***** This happens after a template has been deserialized,
		'***** loads map layers
		Dim oLayerCol As Collection(Of Layer)
		ActiveLayoutMap.Layers.Clear()
		oLayerCol = oMap.GetAllNonGroupLayers
		For Each oLayer As Layer In oLayerCol
			ActiveLayoutMap.Layers.Add(oLayer)
		Next oLayer
		ActiveLayoutMap.MapExtent = oMap.CurrentExtent
		ActiveLayoutMap.BackgroundMask.FillSolidBrush = CType(oMap.BackgroundOverlay.BackgroundBrush, GeoSolidBrush)

		Dim oLegend As CustomLegendLayer = mGetActiveLayoutLegend()

		If Not oLegend Is Nothing Then
			RefreshLegendLayers(oLayerCol, oLegend, Enums.EnumRefreshBehavior.Refresh)
		End If

	End Sub
	Public Function GetScale() As Double
		Dim oPageLayoutMap As CustomMapPrinterLayer = ActiveLayoutMap
		Return oPageLayoutMap.CurrentScale
	End Function
	Private Function mGetPageBoundingBox() As RectangleShape
		Dim oPrinterPage As CustomPagePrinterLayer = mGetPrinterPage()
		Return oPrinterPage.GetPosition(PrintingUnit.Inch)
	End Function
	Private Function mCreateDefaultLegendItem(ByVal oImageStyle As Style, ByVal oTextStyle As TextStyle, ByVal eItemType As CustomLegendItem.EnumLegendItemType) As CustomLegendItem
		Dim oLegendItem As New CustomLegendItem

		'***** A legend item is an element displayed inside of the legend 
		'***** and consists of some text and, optionally a symbol. Some examples
		'***** of legend items would be the legend Title, the names of the layers
		'***** in the map (with the symbol for that layer displayed), the individual
		'***** values for a value map or ranges for a ranged thematic map. There could
		'***** also be a header and a footer.
		With oLegendItem
			.ImageStyle = oImageStyle
			.ItemType = eItemType
			.TextStyle = oTextStyle
			'***** The horizontal and vertical spacing of each CustomLegendItem is determined
			'***** when the font size is set.
			.FontSize = .TextStyle.Font.Size
		End With

		Return oLegendItem

	End Function

	Private Function mGetPrinterPage() As CustomPagePrinterLayer
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oPrinterPage As CustomPagePrinterLayer = Nothing

		If Not oPrinterOverlay Is Nothing Then
			If oPrinterOverlay.PrinterLayers.Contains(csPageLayerKey) Then
				Return CType(oPrinterOverlay.PrinterLayers(csPageLayerKey), CustomPagePrinterLayer)
			End If
		End If
		' ''QB #2339, Jamie Irwin 2/18/2016, Added some exception/debugging code to help us
		' ''identify the bug of opening old TPLs.
		''Dim iIndex As Integer
		''Dim oCPPL As Object
		''If IsNothing(oPrinterPage) Then
		''   For iIndex = 0 To oPrinterOverlay.PrinterLayers.Count - 1
		''      oCPPL = oPrinterOverlay.PrinterLayers(iIndex)
		''      If TypeOf oCPPL Is CustomPagePrinterLayer Then
		''         oPrinterPage = CType(oCPPL, CustomPagePrinterLayer)
		''         Exit For
		''      End If
		''   Next
		''End If
		Return oPrinterPage
	End Function

	Private Function mGetLegend(sKey As String) As CustomLegendLayer
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oLegend As CustomLegendLayer = Nothing

		If Not oPrinterOverlay Is Nothing Then
			If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
				Return CType(oPrinterOverlay.PrinterLayers(sKey), CustomLegendLayer)
			End If
		End If

		Return oLegend

	End Function
	Private Function mGetActiveLayoutLegend() As CustomLegendLayer
		Dim oPrinterOverlay As PrinterInteractiveOverLay = mGetPrinterOverlay()
		Dim oLegend As CustomLegendLayer = Nothing
		Dim sKey As String = mGetKeyForMapElement(csLegendLayerBaseKey, moActiveLayoutMap)

		If Not oPrinterOverlay Is Nothing Then
			If oPrinterOverlay.PrinterLayers.Contains(sKey) Then
				Return CType(oPrinterOverlay.PrinterLayers(sKey), CustomLegendLayer)
			End If
		End If

		Return oLegend
	End Function
	Private Function GetItemWidth(oLegendItem As CustomLegendItem) As Single
		With oLegendItem
			Return moGraphics.MeasureString(.TextStyle.TextColumnName, DisplayUtilities.FontFromGeoFont(.TextStyle.Font)).Width +
						 .LeftPadding + .RightPadding + .TextLeftPadding + .TextRightPadding + .ImageWidth +
						 .ImageLeftPadding + .ImageRightPadding
		End With
	End Function
	Private Function GetItemHeight(oLegendItem As CustomLegendItem) As Single
		Dim dTextHeight As Single
		Dim dImageHeight As Single

		With oLegendItem
			dTextHeight = moGraphics.MeasureString(.TextStyle.TextColumnName, DisplayUtilities.FontFromGeoFont(.TextStyle.Font)).Height +
						 .TextTopPadding + .TextBottomPadding
			dImageHeight = .ImageHeight + .ImageTopPadding + .ImageBottomPadding
			If dTextHeight > dImageHeight Then
				Return dTextHeight
			Else
				Return dImageHeight
			End If
		End With

	End Function
	Private Function mGetKeyForMapElement(sBaseKey As String, oLayoutMap As CustomMapPrinterLayer) As String
		Dim sKey As String

		If sBaseKey = csDataGridBaseKey Then
			'***** There could be multiple data grids so we need to allow for that
			sKey = sBaseKey & "_" & oLayoutMap.PageLayoutMapNumber.ToString & "_" & DateDiff(DateInterval.Second, #1/1/2000#, Now())
		Else
			'***** The other elements for which we create a key in this manner
			'***** There can only be one of these at any given time
			sKey = sBaseKey & "_" & oLayoutMap.PageLayoutMapNumber.ToString
		End If

		Return sKey

	End Function
	Public Shadows Sub Refresh()
		If Utils.CurrentPlugin IsNot Nothing Then
			'***** Save the current Active View state
			Dim eActiveView As Enums.EnumActiveView = Utils.CurrentPlugin.SilvicsMapControl.ActiveView

			'***** If we are in an Auto Layout set that to the active view
			If LayoutMode = MapUtilities.PageLayoutControl.EnumPageLayoutMode.AutoLayout Then
				Utils.CurrentPlugin.SilvicsMapControl.ActiveView = Enums.EnumActiveView.AutoLayoutView
			End If

			For Each oPrinterLayer As PrinterLayer In mGetPrinterOverlay.PrinterLayers
				If TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
					CType(oPrinterLayer, CustomDataGridPrinterLayer).IsDrawing = True
				End If
			Next oPrinterLayer

			MyBase.Refresh()

			For Each oPrinterLayer As PrinterLayer In mGetPrinterOverlay.PrinterLayers
				If TypeOf oPrinterLayer Is CustomDataGridPrinterLayer Then
					CType(oPrinterLayer, CustomDataGridPrinterLayer).IsDrawing = False
				End If
			Next oPrinterLayer

			'***** Restore the saved active view state
			Utils.CurrentPlugin.SilvicsMapControl.ActiveView = eActiveView
		End If
	End Sub
	Public Shadows Sub Refresh(oOverlay As Overlay)
		If Utils.CurrentPlugin IsNot Nothing Then
			'***** Save the current Active View state
			Dim eActiveView As Enums.EnumActiveView = Utils.CurrentPlugin.SilvicsMapControl.ActiveView

			'***** If we are in an Auto Layout set that to the active view
			If LayoutMode = MapUtilities.PageLayoutControl.EnumPageLayoutMode.AutoLayout Then
				Utils.CurrentPlugin.SilvicsMapControl.ActiveView = Enums.EnumActiveView.AutoLayoutView
			End If

			MyBase.Refresh(oOverlay)
			'***** Restore the saved active view state
			Utils.CurrentPlugin.SilvicsMapControl.ActiveView = eActiveView
		End If

	End Sub

	Private Sub moPrinterWorker_DoWork(sender As Object, e As DoWorkEventArgs) Handles moPrinterWorker.DoWork
		Dim sPdfFilePath As String
		Dim eGeoReferenceOption As frmPrintToPdf.EnumWriteGeoReferenceOption
		Dim ePdfFormat As frmPrintToPdf.EnumPdfFormatOption
		Dim oPrinterSettings As PrinterSettings

		Dim oList As List(Of Object) = CType(e.Argument, List(Of Object))

		Dim oPdfOptionsStruct As StructPdfPrintOptions = CType(oList(0), StructPdfPrintOptions)
		With oPdfOptionsStruct
			sPdfFilePath = .PdfFilename
			eGeoReferenceOption = .PdfGeoReferencingOption
			ePdfFormat = .PdfFormatOption
		End With
		oPrinterSettings = CType(oList(1), PrinterSettings)

		mPrint(oPrinterSettings, sPdfFilePath, eGeoReferenceOption, ePdfFormat)
	End Sub

	Private Sub moPrinterWorker_ProgressChanged(sender As Object, e As ProgressChangedEventArgs) Handles moPrinterWorker.ProgressChanged
		Dim sMessage As String = e.UserState.ToString
		If moProgressForm.Cancel Then
			moPrinterWorker.CancelAsync()
		Else
			moProgressForm.SetProgressBarValue(e.ProgressPercentage)
			moProgressForm.UserMessage = sMessage
			moProgressForm.Refresh()
		End If
	End Sub

	Private Sub moPrinterWorker_RunWorkerCompleted(sender As Object, e As RunWorkerCompletedEventArgs) Handles moPrinterWorker.RunWorkerCompleted
		Utils.CurrentPlugin.SilvicsMapControl.ActiveView = meCurrentActiveView
		moProgressForm.Close()
		moProgressForm.Dispose()
		moPrinterWorker.Dispose()
		moPrinterWorker = Nothing
		Refresh()
	End Sub

	Public Sub FlashShape(oShape As BaseShape)
		DisplayUtilities.FlashShape(oShape, Me)
	End Sub
	Protected Overrides Sub OnMouseDown(e As MouseEventArgs)
		MyBase.OnMouseDown(e)
	End Sub
	Protected Overrides Sub OnMouseUp(e As MouseEventArgs)
		MyBase.OnMouseUp(e)
	End Sub
	Private Sub mSimulateMouseUpDown()
		Dim oArgs As New System.Windows.Forms.MouseEventArgs(Windows.Forms.MouseButtons.Middle, 1, 0, 0, 0)
		OnMouseDown(oArgs)
		OnMouseUp(oArgs)
	End Sub

	Private Sub SilvicsPageLayout_MouseWheel(sender As Object, e As MouseEventArgs) Handles Me.MouseWheel
		mSimulateMouseUpDown()
	End Sub

	Private Sub SilvicsPageLayout_Disposed(sender As Object, e As EventArgs) Handles Me.Disposed
		If moGraphics IsNot Nothing Then
			moGraphics.Dispose()
		End If
	End Sub
End Class
